# math
from math import pi
# scipy
from scipy.special import zeta


# PHYSICAL CONSTANTS ################################################

# The fine-structure constant
alpha = 1./137.036

# The electron mass (in MeV)
me = 0.511
# -->
me2 = me**2.

# The classical electron radius (in 1/MeV)
re = alpha/me

# The gravitational constant (in 1/MeV^2)
GN = 6.70861e-45

# The reduced Planck constant (in MeV*s)
hbar = 6.582119514e-22

# The speed of light (in cm/s)
c_si = 2.99792458e10

# The neutron lifetime (in s)
tau_n = 8.794e2 # pre PDG2020: 8.802e2

# The lifetime of tritium (in s)
# T_(1/2) = 3.885e8
tau_t = 5.605e8


# MATHEMATICAL CONSTANTS ############################################

# The Riemann-Zeta function at point 3
zeta3 = zeta(3.)

# The beautiful value of pi, squared
pi2 = pi**2.


# INTERPOLATION-SPECIFIC PARAMETERS #################################

# Boundery values
Emin_log, Tmin_log = 0, -6
Emax_log, Tmax_log = 3, -1
# Number of entries...
num_pd = 150 # ...per decade
Enum = (Emax_log - Emin_log)*num_pd
Tnum = (Tmax_log - Tmin_log)*num_pd


# ALGORITHM-SPECIFIC PARAMETERS #####################################

# The number of nuclei that are
# considered in the Boltzmann equation
# for non-thermal nucleosynthesis
# Default: 9
NY = 9

# The number of mandatory columns in
# 'cosmo_file.dat'
# Default: 5
NC = 5

# The minimum energy for the different spectra (in MeV)
# This value should not be larger than the minimal
# nucleon-interaction threshold of 1.586627 MeV
# (reaction_id: 15 in 'astro-ph/0211258')
# Default: 1.5
Emin = 1.5

# The value that is used for 'approximately' zero
# Default: 1e-200
approx_zero = 1e-200

# The relative accuracy for each integral
# Default: 1e-3
eps = 1e-3

# The maximal value of x = Ephb/T to avoid
# overflow in the exponential function
# Default: 200
Ephb_T_max = 200.

# The maximal value of E/EC up to which the
# integration is performed when using the
# full spectrum with exponential suppression
E_EC_max = 10.

# The number of points per decade for
# the energy grid, which is used within
# the solution of the cascade equation
# Default: 150
NE_pd  = 150
# The minimal number of points for
# the energy grid
# Default: 10
NE_min = 10

# The number of points per decade for
# the temperature grid, which us used
# for the interpolation of the thermal
# nuclear rates
# Default: 50
NT_pd = 50



# The current version of ACROPOLIS
version     = "1.3.0"

# The current dev version of ACROPOLIS
dev_version = "1.3.0"

# The short description of ACROPOLIS
description = "A generiC fRamework fOr Photodisintegration Of LIght elementS"

# The main webpage of ACROPOLIS
url         = "https://acropolis.hepforge.org"

# The authors of ACROPOLIS
authors     = "Paul Frederik Depta, Marco Hufnagel, Kai Schmidt-Hoberg"



# sys
from sys import stdout, stderr, exit

# flags
import acropolis.flags as flags
# info
from acropolis.info import version, dev_version, url


_max_verbose_level = 1

_use_color = True


def print_version():
    if flags.verbose:
        # Differentiate between stable and dev version
        version_str = ""
        # Stable version
        if version == dev_version:
            version_str = "v{}".format(version)
        # Development version
        else:
            version_str = "v{} [dev]".format(dev_version)

        if _use_color:
            ctxt = "\x1B[38;5;209m"
            cend = "\x1B[0m"
        else:
            ctxt = cend = ""

        stdout.write( f"{ctxt}ACROPOLIS {version_str} ({url}){cend}\n\n" )


def print_Yf(Yf, header=["mean", "high", "low"]):
    # If not verbose, simply print one line
    # including all abundances
    if not flags.verbose:
        print(*Yf.transpose().reshape(1, Yf.size)[0,:])
        return

    # Define the colors
    if _use_color:
        chdr = "\x1B[35m"
        celm = "\x1B[34m"
        cdcy = "\x1B[36m"
        cend = "\x1B[0m"
    else:
        chdr = celm = cdcy = cend = ""

    # Fill potentially missing header entries
    NYf = Yf.shape[1]
    header.extend( [""] * ( NYf - len(header) ) )

    # Set small values to zero to guarantee the same
    # width for all abundances in the output
    Yf[Yf <= 1e-99] = 0

    # Define the set of all possible labels
    labels = ['n', 'p', 'H2', 'H3', 'He3', 'He4', 'Li6', 'Li7', 'Be7']

    # Print the header
    header_str = "\n{:^4}"
    for i in range(NYf):
        header_str  += f" | {chdr}    {{:8}}{cend}"

    print( header_str.format("", *header) )
    print("-------------------------------------------------")

    # Print the different abundances
    for j, label in enumerate(labels):
        line = f"{celm}{{:>4}}{cend}"
        for i in range(NYf):
            line += " | {:11.6e}"

        if label in ['n', 'H3', 'Be7']:
            line += f"  [{cdcy}{{:7}}{cend}]"

        print( line.format(label, *Yf[j], 'decayed') )


def print_error(error, loc="", eol="\n", flush=False):
    # Define the colors
    if _use_color:
        cloc  = "\x1B[1;35m"
        ctyp  = "\x1B[1;31m"
        cend  = "\x1B[0m"
    else:
        cloc = ctyp = cend = ""

    locf = ""
    if flags.debug and loc != "":
        locf = f" {cloc}({loc}){cend}"

    stderr.write(f"{ctyp}ERROR  {cend}: {error}{locf}{eol}")
    
    if flush:
        stderr.flush()
    
    exit(1)


def print_warning(warning, loc="", eol="\n", flush=False):
    # Define the colors
    if _use_color:
        cloc  = "\x1B[1;35m"
        ctyp = "\x1B[1;33m"
        cend  = "\x1B[0m"
    else:
        cloc = ctyp = cend = ""
    
    locf = ""
    if flags.debug and loc != "":
        locf = f" {cloc}({loc}){cend}"

    stdout.write(f"{ctyp}WARNING{cend}: {warning}{locf}{eol}")
    
    if flush:
        stdout.flush()


def print_info(info, loc="", eol="\n", flush=False, verbose_level=None):
    global _max_verbose_level

    if verbose_level is None:
        verbose_level = _max_verbose_level

    _max_verbose_level = max( _max_verbose_level, verbose_level )

    # Define the colors
    if _use_color:
        cloc  = "\x1B[1;35m"
        ctyp = "\x1B[1;32m"
        cend  = "\x1B[0m"
    else:
        cloc = ctyp = cend = ""

    locf = ""
    if flags.debug and loc != "":
        locf = f" {cloc}({loc}){cend}"

    if flags.verbose and verbose_level >= _max_verbose_level:
        stdout.write(f"{ctyp}INFO   {cend}: {info}{locf}{eol}")
        
        if flush:
            stdout.flush()


def set_max_verbose_level(max_verbose_level=None):
    global _max_verbose_level

    if max_verbose_level is None:
        max_verbose_level = 1

    _max_verbose_level = max_verbose_level


def disable_color():
    global _use_color

    _use_color = False


# ATTENTION !!!!!111elf
# Only parameters that specify a default value are
# meant to be changed by the user, i.e. everything
# under FLAGS and ALGORITHM-SPECIFIC PARAMETERS


# FLAGS #############################################################

# If this flag is set to 'True',
# the pregenerated databases
# will be used to interpolate
# the different reaction rates
# Default: True
usedb = True

# If this flag is set to 'True',
# additional output is printed
# to the screen
# Default: True
verbose = True

# If this flag is set to 'True',
# additional debug information
# is printed to the screen
# Default: False
debug = False

# If this flag is set to 'True',
# the universal spectrum is used
# for all points in parameter space
# ATTENTION:
# Change with caution and only if
# you know what you are doing.
# World destruction possible!
# Default: False
universal = False


def set_universal(value):
    global universal

    universal = value


# math
from math import pi, exp, log, log10
# numpy
import numpy as np
# scipy
from scipy.linalg import expm
# abc
from abc import ABC, abstractmethod

# input
from acropolis.input import InputInterface, locate_sm_file
# nucl
from acropolis.nucl import NuclearReactor, MatrixGenerator
# params
from acropolis.params import zeta3
from acropolis.params import hbar, c_si, me2, alpha, tau_t
from acropolis.params import Emin, NY
# flags
import acropolis.flags as flags
# pprint
from acropolis.pprint import print_info, print_warning


class AbstractModel(ABC):

    def __init__(self, e0, ii):
        # Initialize the input interface
        self._sII  = ii

        # The injection energy
        self._sE0 = e0

        # The temperature range that is used for the calculation
        self._sTrg = self._temperature_range()

        # The relevant source terms
        (self._sS0, self._sSc) = self.get_source_terms()

        # A buffer for high-performance scans
        self._sMatpBuffer = None


    def run_disintegration(self):
        # Print a warning if the injection energy
        # is larger than 1GeV, as this might lead
        # to wrong results
        if not flags.universal and int( self._sE0 ) > 1e3:
            print_warning(
                "Injection energy > 1 GeV. Results cannot be trusted.",
                "acropolis.models.AbstractMode.run_disintegration"
            )

        # Print a warning if the temperature range
        # of the model is not covered by the data
        # in cosmo_file.dat
        cf_temp_rg = self._sII.temperature_range()
        if not (cf_temp_rg[0] <= self._sTrg[0] <= self._sTrg[1] <= cf_temp_rg[1]):
            print_warning(
                "Temperature range not covered by input data. Results cannot be trusted.",
                "acropolis.models.AbstractMode.run_disintegration"
            )

        # If the energy is below all thresholds,
        # simply return the initial abundances
        if self._sE0 <= Emin:
            print_info(
                "Injection energy is below all thresholds. No calculation required.",
                "acropolis.models.AbstractModel.run_disintegration",
                verbose_level=1
            )
            return self._squeeze_decays( self._sII.bbn_abundances() )

        # Calculate the different transfer matrices
        ###########################################

        # 1. pre-decay
        pred_mat   = self._pred_matrix()
        # 2. photodisintegration
        pdi_mat    = self._pdi_matrix()
        # 3. post-decay
        postd_mat  = self._postd_matrix()

        # Combine
        transf_mat = postd_mat.dot( pdi_mat.dot( pred_mat ) )

        # Calculate the final abundances
        Yf = np.column_stack(
            list( transf_mat.dot( Y0i ) for Y0i in self._sII.bbn_abundances().transpose() )
        )

        return Yf


    def get_source_terms(self):
        # Collect the different source terms, i.e. ...
        # ...the 'delta' source terms and...
        s0 = [
            self._source_photon_0  ,
            self._source_electron_0,
            self._source_positron_0
        ]
        # ...the continous source terms
        sc = [
            self._source_photon_c  ,
            self._source_electron_c,
            self._source_positron_c
        ]

        return (s0, sc)


    def _pdi_matrix(self):
        if self._sMatpBuffer is None:
            # Initialize the NuclearReactor
            nr = NuclearReactor(self._sS0, self._sSc, self._sTrg, self._sE0, self._sII)

            # Calculate the thermal rates
            (temp, pdi_grids) = nr.get_pdi_grids()

            # Initialize the MatrixGenerator
            mg = MatrixGenerator(temp, pdi_grids, self._sII)

            # Calculate the final matrices and set the buffer
            self._sMatpBuffer = mg.get_final_matp()

        # Calculate the final matrices
        matp = self._sMatpBuffer

        # Calculate the final matrix and return
        return expm( sum(m for m in matp) )


    def _pred_matrix(self):
        dmat = np.identity(NY)

        # n   > p
        dmat[0,0], dmat[1,0] = 0., 1.

        # t > He3
        Tmax = max( self._sTrg )
        tmax = self._sII.time( Tmax )

        expf = exp( -tmax/tau_t )

        dmat[3,3], dmat[4, 3] = expf, 1. - expf

        return dmat


    def _postd_matrix(self):
        dmat = np.identity(NY)

        dmat[0,0], dmat[1,0] = 0., 1. # n   > p
        dmat[3,3], dmat[4,3] = 0., 1. # t   > He3
        dmat[8,8], dmat[7,8] = 0., 1. # Be7 > Li7

        return dmat


    def _squeeze_decays(self, Yf):
        dmat = self._postd_matrix()

        return np.column_stack(
            list( dmat.dot( Yi ) for Yi in Yf.transpose() )
        )


    def get_matp_buffer(self):
        return self._sMatpBuffer


    def set_matp_buffer(self, matp):
        self._sMatpBuffer = matp


    # ABSTRACT METHODS ##############################################

    @abstractmethod
    def _temperature_range(self):
        pass


    @abstractmethod
    def _source_photon_0(self, T):
        pass


    @abstractmethod
    def _source_electron_0(self, T):
        pass


    def _source_positron_0(self, T):
        return self._source_electron_0(T)


    def _source_photon_c(self, E, T):
        return 0.


    def _source_electron_c(self, E, T):
        return 0.


    def _source_positron_c(self, E, T):
        return self._source_electron_c(E, T)


class DecayModel(AbstractModel):

    def __init__(self, mphi, tau, temp0, n0a, bree, braa):
        # Initialize the Input_Interface
        self._sII   = InputInterface( locate_sm_file() )

        # The mass of the decaying particle
        self._sMphi = mphi            # in MeV
        # The lifetime of the decaying particle
        self._sTau  = tau             # in s
        # The injection energy
        self._sE0   = self._sMphi/2.  # in MeV

        # The number density of the mediator
        # (relative to photons) ...
        self._sN0a  = n0a
        # ... at T = temp0 ...
        self._sT0   = temp0           # in MeV
        # ... corresponding to t = t(temp0)
        self._st0   = self._sII.time(self._sT0)

        # The branching ratio into electron-positron pairs
        self._sBRee = bree
        # The branching ratio into two photons
        self._sBRaa = braa

        # Call the super constructor
        super(DecayModel, self).__init__(self._sE0, self._sII)


    # DEPENDENT QUANTITIES ##############################################################

    def _number_density(self, T):
        sf_ratio = self._sII.scale_factor(self._sT0)/self._sII.scale_factor(T)

        delta_t = self._sII.time(T) - self._st0
        n_gamma = (2.*zeta3)*(self._sT0**3.)/(pi**2.)

        return self._sN0a * n_gamma * sf_ratio**3. * exp( -delta_t/self._sTau )


    # ABSTRACT METHODS ##################################################################

    def _temperature_range(self):
        # The number of degrees-of-freedom to span
        mag = 2.
        # Calculate the approximate decay temperature
        Td = self._sII.temperature( self._sTau )
        # Calculate Tmin and Tmax from Td
        Td_ofm = log10(Td)
        # Here we choose -1.5 (+0.5) orders of magnitude
        # below (above) the approx. decay temperature,
        # since the main part happens after t = \tau
        Tmin = 10.**(Td_ofm - 3.*mag/4.)
        Tmax = 10.**(Td_ofm + 1.*mag/4.)

        return (Tmin, Tmax)


    def _source_photon_0(self, T):
        return self._sBRaa * 2. * self._number_density(T) * (hbar/self._sTau)


    def _source_electron_0(self, T):
        return self._sBRee * self._number_density(T) * (hbar/self._sTau)


    def _source_photon_c(self, E, T):
        EX = self._sE0

        x = E/EX
        y = me2/(4.*EX**2.)

        if 1. - y < x:
            return 0.

        _sp = self._source_electron_0(T)

        return (_sp/EX) * (alpha/pi) * ( 1. + (1.-x)**2. )/x * log( (1.-x)/y )


class AnnihilationModel(AbstractModel):

    def __init__(self, mchi, a, b, tempkd, bree, braa, omegah2=0.12):
        # Initialize the Input_Interface
        self._sII    = InputInterface( locate_sm_file() )

        # The mass of the dark-matter particle
        self._sMchi  = mchi         # in MeV
        # The s-wave and p-wave parts of <sigma v>
        self._sSwave = a            # in cm^3/s
        self._sPwave = b            # in cm^3/s
        # The dark-matter decoupling temperature. For
        # Tkd=0, the dark matter partices stays in
        # kinetic equilibrium with the SM heat bath
        self._sTkd   = tempkd       # in MeV
        # The injection energy
        self._sE0    = self._sMchi  # in MeV

        # The branching ratio into electron-positron pairs
        self._sBRee  = bree
        # The branching ratio into two photons
        self._sBRaa  = braa

        # The density parameter of dark matter
        self._sOmgh2 = omegah2

        # Call the super constructor
        super(AnnihilationModel, self).__init__(self._sE0, self._sII)


    # DEPENDENT QUANTITIES ##############################################################

    def _number_density(self, T):
        rho_d0 = 8.095894680377574e-35 * self._sOmgh2  # DM density today in MeV^4
        T0     = 2.72548*8.6173324e-11                 # CMB temperature today in MeV

        # Calculate the scale factor at T0 via extrapolation
        Tmin = self._sII.temperature_range()[0] * ( 1 + 1e-6 )
        # -->
        sf0 = self._sII.scale_factor(Tmin) * Tmin / T0

        sf_ratio = sf0 / self._sII.scale_factor(T)

        return rho_d0 * sf_ratio**3. / self._sMchi


    def _dm_temperature(self, T):
        if T >= self._sTkd:
            return T

        sf_ratio = self._sII.scale_factor(self._sTkd) / self._sII.scale_factor(T)

        return self._sTkd * sf_ratio**2.


    def _sigma_v(self, T):
        swave_nu = self._sSwave/( (hbar**2.)*(c_si**3.) )
        pwave_nu = self._sPwave/( (hbar**2.)*(c_si**3.) )

        v2 = 6.*self._dm_temperature(T)/self._sMchi

        return swave_nu + pwave_nu*v2


    # ABSTRACT METHODS ##################################################################

    def _temperature_range(self):
        # The number of degrees-of-freedom to span
        mag = 4.
        # Tmax is determined by the Be7 threshold
        # (The factor 0.5 takes into account effects
        # of the high-energy tail)
        Tmax = me2/(22.*.5*Emin)
        # For smaller T the annihilation rate is suppressed
        # --> falls off at least with T^(-6)
        Tmin = 10.**(log10(Tmax) - mag)

        return (Tmin, Tmax)


    def _source_photon_0(self, T):
        return self._sBRaa * (self._number_density(T)**2.) * self._sigma_v(T)


    def _source_electron_0(self, T):
        return self._sBRee * .5 * (self._number_density(T)**2.) * self._sigma_v(T)


    def _source_photon_c(self, E, T):
        EX = self._sE0

        x = E/EX
        y = me2/(4.*EX**2.)

        if 1. - y < x:
            return 0.

        _sp = self._source_electron_0(T)

        return (_sp/EX) * (alpha/pi) * ( 1. + (1.-x)**2. )/x * log( (1.-x)/y )


# math
from math import sqrt, exp, log
# numpy
import numpy as np
# scipy
from scipy.integrate import quad
from scipy.optimize import root
from scipy.special import kn

# models
from acropolis.models import AnnihilationModel
# params
from acropolis.params import me
from acropolis.params import pi, pi2, zeta3
from acropolis.params import eps
# pprint
from acropolis.pprint import print_error


# https://stackoverflow.com/questions/1167617/in-python-how-do-i-indicate-im-overriding-a-method
def overrides(interface_class):
    def overrider(method):
        assert(method.__name__ in dir(interface_class))
        
        return method

    return overrider


def estimate_tempkd_ee(mchi, delta, gammad, gammav, nd, S, ii, sigma_ee):
    fac = 200

    # The integration kernel for calculating the
    # thermally averaged cross-section for
    # \chi e^\pm -> \chi e^\pm scattering
    def _sigma_v_ee_kernel(logz, T):
        z = exp(logz)
        # -->
        s = (z + me + mchi)**2. # sqrt_s = z + me + mchi
        # -->
        sqrt_s = sqrt(s)

        sigma = sigma_ee(s=s, mchi=mchi, delta=delta, gammad=gammad, gammav=gammav)

        bessel_term = 0

        def _k1_apx_wo_exp(y):
            return sqrt(pi/2) * ( y**0.5 +  (3./8.) * y**1.5 -  (15./128.) * y**2.5 + (105./1024.) * y**3.5)
        
        def _k2_apx_wo_exp(y):
            return sqrt(pi/2) * ( y**0.5 + (15./8.) * y**1.5 + (105./128.) * y**2.5 - (315./1024.) * y**3.5)

        # Define the cutoff for when to Taylor expand the Bessel functions
        xmax = 400

        ye, ychi, ys = T/me, T/mchi, T/sqrt_s
        # For mchi/T > xmax, we have sqrt_s/T > (me+mchi)/T > xmax
        if me/T < xmax < mchi/T:
            # Expand K1(sqrt_s/T) and K2(mchi/T)
            # z + me = sqrt_s - mchi
            bessel_term = exp( -(z + me)/T ) * _k1_apx_wo_exp(ys) / _k2_apx_wo_exp(ychi) / kn(2, me/T)
        elif me/T > xmax:
            # Expand K1(sqrt_s/T), K2(mchi/T) and K2(me/T)
            # z = sqrt_s - me - mchi
            bessel_term = exp( -z/T ) * _k1_apx_wo_exp(ys) / ( _k2_apx_wo_exp(ychi) * _k2_apx_wo_exp(ye) )
        else:
            # Expand nothing
            bessel_term = kn(1, sqrt_s/T) / ( kn(2, me/T) * kn(2, mchi/T) )

        # s = (z + me + mchi)^2 => ds/dz = 2(z + me + mchi) = 2 sqrt_s 
        return z * (2.*sqrt_s) * sigma * ( s - (me + mchi)**2. ) * ( s - (me - mchi)**2. ) * bessel_term / sqrt_s


    # The thermally averaged cross-section for
    # \chi e^\pm -> \chi e^\pm scattering
    def _sigma_v_ee(T):
        zmin = T/fac
        zmax = fac*T

        integral = quad(_sigma_v_ee_kernel, log(zmin), log(zmax), epsrel=eps, epsabs=0, limit=100, args=(T,))

        return integral[0]/( 8. * me**2. * mchi**2. * T)


    # The number density of e^\pm
    def _nee(T):
        xe = me/T

        # 1. EQUILIBRIUM WITH VANISHING CHEM. POTENTIAL
        f_kernel = lambda y, x: y * sqrt(y**2. - x**2.) / ( exp(y) + 1 ) if x <= y else 0. # x = me/T, y = Ee/T
        # -->
        f = quad(f_kernel, xe, fac, epsabs=0, epsrel=eps, args=(xe,))[0]
        # -->
        nee_1 = 4. * f * (T**3.) / ( 2. * pi2 )
        # ne = ge * T^3 * quad(f) / 2 pi^2

        # 1. EQUILIBRIUM WITH NON-VANISHING CHEM. POTENTIAL
        Y = ii.bbn_abundances_0()
        # -->
        nee_2 = ( Y[1] + 2.*Y[5] ) * ii.parameter('eta') * ( 2.*zeta3/pi2 ) * (T**3.)

        return max(nee_1, nee_2)


    # Ncol * H ~ sig_v * _nee
    def _tempkd_ee_root(logT):
        T = exp( logT )

        Ncol = max(1., mchi/T)

        return log( _sigma_v_ee(T) * _nee(T) / ii.hubble_rate(T) / Ncol )
    
    # -->
    tempkd = exp( root(_tempkd_ee_root, 0.).x )
    
    # Assume that kinetic decoupling happens at least
    # together with chemical decoupling at T ~ mchi/20
    return min(tempkd, mchi/20.)


# This model has been created in collaboration with Pieter Braat (pbraat@nikhef.nl)
# When using this model, please cite arXiv:2406:XXXX
class ResonanceModel(AnnihilationModel):
    def __init__(self, mchi, delta, gammad, gammav, nd, tempkd, S=1, omegah2=0.12):

        # CALL THE SUPER CONSTRUCTOR (OF ANNIHILATION_MODEL) ##################
        #######################################################################

        super(ResonanceModel, self).__init__(
        #   mchi, a   , b   , tempkd, bree, braa, omegah2
            mchi, None, None, None  , 1   , 0   , omegah2
        #                     | tempkd will be set below
        )

        # CALCULATE TKD FROM THE GIVEN FUNCTION IF REQUESTED ##################
        #######################################################################
        if callable(tempkd):
            self._sTkd = tempkd(
                mchi=mchi, delta=delta, gammad=gammad, gammav=gammav, nd=nd, S=S, ii=self._sII
            )
        else:
            self._sTkd = tempkd


        # SPECIFY THE NEW PARAMETERS ##########################################
        #######################################################################

        # The mass splitting between the resonant
        # particle and the dark-matter particle
        # mr = mchi * ( 2 + delta )
        self._sDelta  = delta
        # The couplings to the dark and to the
        # visible sector
        self._sGammad = gammad
        self._sGammav = gammav
        # The parameters to distinguish between
        # s- (nd=0) and p-wave (nd=1) processes
        self._sNd     = nd
        self._sNv     = 1 # NOTE: Only electrons are allowed in the final state for now
        # The symmetry factor for the annihilation cross-section
        self._sS      = S

        # The mass of the resonant particle
        self._sMR     = self._sMchi * ( 2. + self._sDelta ) # in MeV
        # The resonance momentum
        self._sPR     = self._sMchi * sqrt(self._sDelta)    # in MeV
        # The total decay width of the resonant particle
        self._sWidth  = self._decay_width(self._sPR)        # in MeV


        # CHECK THE INPUT PARAMATERS ##########################################
        #######################################################################
        
        self._check_input_parameters()


    def _check_nwa(self, eps=0.1):
        y = self._sWidth / self._sMR
        # In the NWA limit, we have y -> 0, i.e. y < eps

        if y < eps and self._sNd != 0:
            return True

        return False
    

    def _check_input_parameters(self):
        if self._sDelta > 1:
            print_error(
                "The mass splitting must be < 1. The calculation cannot be trusted.",
                "acropolis.models.ResonanceModel._check_input_parameters"
            )
        
        if (self._sGammad >= 1 or self._sGammav >= 1): 
            print_error(
                "The couplings must be small (< 1). The calculation cannot be trusted.",
                "acropolis.model.ResonanceModel._check_input_parameters"
            )
        
        if self._sNd not in [0, 1]:
            print_error(
                "Currently only s-wave annihilations with 'nd = 0' and p-wave " + \
                "annihilations with 'nd = 1' are supported.",
                "acropolis.models.ResonanceModel._check_input_parameters"
            )


    # DEPENDENT QUANTITIES ##############################################################


    # The total decay width of the resonant 
    # particle into dark-sector states
    def _decay_width_d(self, p):
        return self._sGammad * self._sMR * (p / self._sMchi)**(2.*self._sNd + 1.)
    

    # The total decay width of the resonant 
    # particle into visible-sector states
    def _decay_width_v(self, p):
        return self._sGammav * self._sMR * (p / self._sMchi)**(2.*self._sNv + 1.)

    
    # The total decay width of the resonant
    # particle
    def _decay_width(self, p):
        return self._decay_width_d(p) + self._decay_width_v(p)


    # The thermally average annihilation
    # cross-section in the resonant regime
    def _sigma_v_res(self, T):
        x = self._sMchi/T

        return 8. * self._sS * (pi*x)**1.5 * self._sGammad * self._sGammav * self._sMR**2. * self._sDelta**(self._sNd+.5) * exp(-self._sDelta*x) \
                / self._sWidth / self._sMchi**3.

    
    # The thermally averaged annihilation
    # cross-section in the non-resonant regime
    def _sigma_v_non_res(self, T):
        x = self._sMchi/T

        # Speed up the calculation: only nd = 0, 1 are allowed
        # gamma(self._sNd+1.5)
        gamma = {
            0:      sqrt(pi) / 2., # \gamma(1.5)
            1: 3. * sqrt(pi) / 4.  # \gamma(2.5)
        }[self._sNd]

        return 4. * self._sS * sqrt(pi) * x**(-self._sNd) * self._sGammad * self._sGammav * self._sMR**2. * gamma \
                / ( self._sMchi**4. * self._sDelta**2. )
    

    # The full (non-approximate) thermally
    # averaged annihilation cross section
    def _sigma_v_full(self, T):
        x  = self._sMchi/T

        uR = self._sPR**2.
        m2 = self._sMchi**2.

        # Set the maximal exponent in the integral
        # beyond which the integrand is cut off
        exp_cutoff = 200.

        # Define the prefactor
        pref = 4. * x * sqrt(x*pi) * self._sS * self._sGammad * self._sGammav * (self._sMR/self._sMchi)**2.    

        # Define the integration kernel
        def _sigma_v_full_kernel(log_u):
            u      = exp(log_u) # = p^2
            sqrt_u = sqrt(u)    # = p

            width_u = self._decay_width(sqrt_u)

            #      | from log integration
            return u * exp( -u*x/m2 ) * ( u/m2 )**(self._sNd+.5) / ( (u - uR)**2. + ( self._sMchi*width_u/2. )**2. )

        # Calculate the upper integration limit
        umax = exp_cutoff * m2 / x

        uR_l, uR_h = uR*(1-eps), uR*(1+eps) 

        # uR_h/uR_lt ~ 1.002
        # This is still good with exp_cutoff
        uR_l = min(uR_l, umax)

        # Perform the integration in three steps
        # BELOW RESONANCE
        I1 = quad(
            _sigma_v_full_kernel, -np.inf  , log(uR_l), epsrel=eps, epsabs=0
        )[0]
        # AROUND RESONANCE
        I2 = quad(
            _sigma_v_full_kernel, log(uR_l), log(uR_h), epsrel=eps, epsabs=0, limit=100, points=(log(uR),)
        )[0] if uR_l < umax else 0.
        # ABOVE RESONANCE
        I3 = quad(
            _sigma_v_full_kernel, log(uR_h), log(umax), epsrel=eps, epsabs=0
        )[0] if uR_l < umax else 0.

        return pref*(I1+I2+I3)


    @overrides(AnnihilationModel)
    def _sigma_v(self, T):
        return self._sigma_v_full(
            self._dm_temperature(T)
        )


# functools
from functools import wraps
# math
from math import sqrt, log10, log, exp
# numpy
import numpy as np
# scipy
from scipy.integrate import quad
from scipy.integrate import IntegrationWarning
# time
from time import time
# warnings
import warnings

# util
from acropolis.utils import LogInterp
# pprint
from acropolis.pprint import print_error, print_warning, print_info
# params
from acropolis.params import me2, hbar, tau_n, tau_t
from acropolis.params import approx_zero, eps, E_EC_max
from acropolis.params import NT_pd, NY
# flags
import acropolis.flags as flags
# cascade
from acropolis.cascade import SpectrumGenerator


# A dictionary containing all relevant nuclei, or more precisely
# all nuclei that appear in the reactions specified in '_reactions'
# or in the decays specified in '_decays'
_nuclei = {
    "n"  : 0,
    "p"  : 1,
    "d"  : 2,
    "t"  : 3,
    "He3": 4,
    "He4": 5,
    "Li6": 6,
    "Li7": 7,
    "Be7": 8
}


# A dictionary containing all relevant pdi reactions
# This dict can be modified if new reactions are added
# In this case, also remember to modify the function
# 'NuclearReactor.get_cross_section(reaction_id, E)'
_reactions = {
    1 : "d+a>n+p",
    2 : "t+a>n+d",
    3 : "t+a>n+p+n",
    4 : "He3+a>p+d",
    5 : "He3+a>n+p+p",
    6 : "He4+a>p+t",
    7 : "He4+a>n+He3",
    8 : "He4+a>d+d",
    9 : "He4+a>n+p+d",
    10: "Li6+a>n+p+He4",
    11: "Li6+a>X",
    12: "Li7+a>t+He4",
    13: "Li7+a>n+Li6",
    14: "Li7+a>n+n+p+He4",
    15: "Be7+a>He3+He4",
    16: "Be7+a>p+Li6",
    17: "Be7+a>p+p+n+He4"
}
# A dictionary containing all accociated threshold
# energies. All energies are given in MeV
_eth = {
    1 :  2.224573,
    2 :  6.257248,
    3 :  8.481821,
    4 :  5.493485,
    5 :  7.718058,
    6 : 19.813852,
    7 : 20.577615,
    8 : 23.846527,
    9 : 26.071100,
    10:  3.698892,
    11: 15.794685,
    12:  2.467032,
    13:  7.249962,
    14: 10.948850,
    15:  1.586627,
    16:  5.605794,
    17:  9.304680
}
# A dictionary containing the theoretical errors for
# the different reaction rates (taken from 2006.14803)
# in terms of a relative deviation from the mean value
# 8 (He4->d+d); 10, 11; (Li6->...); 12, 14 (Li7->...)
# TODO: Perform literature search and find sensible values
_rdev = {
    1:  0.00,
    2:  0.00,
    3:  0.00,
    4:  0.00,
    5:  0.00,
    6:  0.00,
    7:  0.00,
    8:  0.00,
    9:  0.00,
    10: 0.00,
    11: 0.00,
    12: 0.00,
    13: 0.00,
    14: 0.00,
    15: 0.00,
    16: 0.00,
    17: 0.00
}


# A dictionary containing all relevant decays
_decays = {
    1: "n>p",
    2: "t>He3"
    #3: "Be7>Li7"
}
# A dictionary containing all accociated lifetimes.
# All lifetimes are given in s
_tau = {
    1: tau_n,
    2: tau_t
    #3: 6.634e6         # T_(1/2) = 4.598e6
}


# The number of relevant nucleons
_nnuc = len( _nuclei )
# The number of relevant reaction
_nrec = len( _reactions )
# The number of relevant decays
_ndec = len( _decays )


# A list containing all reactions id's
_lrid = list( _reactions.keys() )
# A list containing all decay id's
_ldid = list( _decays.keys() )


def _extract_signature(reaction_str):
    sp = reaction_str.split(">")

    # Extract the inital state
    istate = _nuclei[ sp[0].split("+")[0] ]

    # Extract the final state
    #
    # Set up a default dictionary (to store the final nucleons)
    fstate = { i:0 for i in range( _nnuc ) }

    # Fill the dictionary; Afterwards this variable stores
    # the number of all nuleids in the final state
    # The result looks somewhat like
    # {0:<number_of_n_in_fs>, 1:<number_of_p_in_fs, 3:...}
    for Nstr in sp[1].split("+"):
        if Nstr in _nuclei: # Do not consider particles like X in reaction 11
            fstate[ _nuclei[Nstr] ] += 1

    return istate, fstate


# A dictionary containing the signatures
# for the different pdi reactions
_rsig = { rid:_extract_signature( _reactions[rid] ) for rid in _lrid }
# A dictionary containing the signatures
# for the different decays
_dsig = { did:_extract_signature( _decays[ did]   ) for did in _ldid }


def _convert_mb_to_iMeV2(f_in_mb):
    # Define the conversion factor
    cf = 2.56819e-6

    # Define the wrapper function
    @wraps(f_in_mb)
    def f_in_iMeV2(*args, **kwargs):
        return cf * f_in_mb(*args, **kwargs)

    return f_in_iMeV2


class NuclearReactor(object):

    def __init__(self, s0, sc, temp_rg, e0, ii):
        self._sII = ii

        # The injection energy
        self._sE0 = e0

        # The prefactor of the delta source term
        self._sS0 = s0

        # The continouos source terms
        self._sSC = sc

        # The approximate decay temperature of the mediator
        self._sTrg  = temp_rg

        # An instance of 'Spectrum_Generator' in order to calculate
        # the photon spectrum in the function 'get_reaction_rate(reaction_id, T)'
        self._sGen = SpectrumGenerator(self._sII)

    # BEGIN REACTIONS ###############################################

    def _generic_expr(self, E, Q, N, p1, p2, p3):
        # Below threshold, return 0
        if E < Q:
            return 0.

        return N * (Q**p1) * (E-Q)**p2 / (E**p3)


    def _da_np(self, E):
        Q = _eth[1]

        # Below threshold, return 0.
        if E < Q:
            return 0.

        return 18.75 * ( ( sqrt( Q*(E-Q) )/E )**3. + 0.007947*( sqrt( Q*(E-Q) )/E )**2. * ( (sqrt(Q) - sqrt(0.037))**2./( E - Q + 0.037 ) ) )


    def _ta_nd(self, E):
        return self._generic_expr(E, _eth[2], 9.8, 1.95, 1.65, 3.6)


    def _ta_npn(self, E):
        return self._generic_expr(E, _eth[3], 26.0, 2.6, 2.3, 4.9)


    def _He3a_pd(self, E):
        return self._generic_expr(E, _eth[4], 8.88, 1.75, 1.65, 3.4)


    def _He3a_npp(self, E):
        return self._generic_expr(E, _eth[5], 16.7, 1.95, 2.3, 4.25)


    def _He4a_pt(self, E):
        return self._generic_expr(E, _eth[6], 19.5, 3.5, 1.0, 4.5)


    def _He4a_nHe3(self, E):
        # Prefactor changed from 17.1mb to 20.7mb to account for EXFOR data
        # cf. 'hep-ph/0604251 [38]' for more details
        return self._generic_expr(E, _eth[7], 20.7, 3.5, 1.0, 4.5)


    def _He4a_dd(self, E):
        return self._generic_expr(E, _eth[8], 10.7, 10.2, 3.4, 13.6)


    def _He4a_npd(self, E):
        return self._generic_expr(E, _eth[9], 21.7, 4.0, 3.0, 7.0)


    def _Li6a_npHe4(self, E):
        # Prefactor changed from 104.0mb to 143.0mb to account for EXFOR data
        # cf. 'hep-ph/0604251 [39]' for more details
        return self._generic_expr(E, _eth[10], 143.0, 2.3, 4.7, 7.0)


    def _Li6a_XA3(self, E):
        Q = _eth[11]

        # Template for the exponential term of the form
        # -- N * exp( -1/2*( (E - Eb)/Ed )^2 ) --
        # E, Eb, Ed in MeV, N unitless, result unitless
        def exp_term(E, N, Eb, Ed):
            return N * exp( -(1./2.)*( (E - Eb)/Ed )**2. )

        # __genereic_expr returns 0. below threshold
        return self._generic_expr(E, Q, 38.1, 3.0, 2.0, 5.0) * ( exp_term(E, 3.7, 19.0, 3.5) + exp_term(E, 2.75, 30.0, 3.0) + exp_term(E, 2.2, 43.0, 5.0) )


    def _Li7a_tHe4(self, E):
        Q = _eth[12]

        # Below threshold, return 0.
        if E < Q:
            return 0.

        # Define the excess energy relative to the threshold
        Ecm = E - Q

        # Define the closing polynomial in Ecm...
        pEcm = 1. + 2.2875*(Ecm**2.) - 1.1798*(Ecm**3.) + 2.5279*(Ecm**4.)

        # ...and for pEcm < 0, return 0. (This should be a continuous transition)
        # For this reaction, however, there should not be any problems
        # The roots of pEcm are all imaginary according to 'WolframAlpha'
        # Therefore, do not perform the check in order not to loose performance
        #if pEcm < 0.:
        #    return 0.

        return 0.105 * ( 2371./(E**2) ) * exp( -2.5954/sqrt(Ecm) ) * exp(-2.056*Ecm) * pEcm


    def _Li7a_nLi6(self, E):
        Q = _eth[13]

        # Below threshold, return 0.
        if E < Q:
            return 0.

        return self._generic_expr(E, Q, 0.176, 1.51, 0.49, 2.0) + self._generic_expr(E, Q, 1205.0, 5.5, 5.0, 10.5) + 0.06/( 1. + ( (E - Q - 7.46)/0.188 )**2. )


    def _Li7a_nnpHe4(self, E):
        return self._generic_expr(E, _eth[14], 122.0, 4.0, 3.0, 7.0)


    def _Be7a_He3He4(self, E):
        Q = _eth[15]

        # Below threshold, return 0.
        if E < Q:
            return 0.

        # Define the excess energy relative to the threshold
        Ecm = E - Q

        # Define the closing polynomial in Ecm...
        pEcm = 1. - 0.428*(Ecm**2.) + 0.534*(Ecm**3.) - 0.115*(Ecm**4.)

        # ... and for pEcm < 0, return 0. (This should be a continuous transition)
        # In fact, pEcm has a root at Ecm ~ 3.92599MeV > Q according to 'WolframAlpha',
        # This root lies above threshold, which is why we have to explicitly perform
        # the corresponding check for this reaction
        if pEcm < 0.:
            return 0.

        return 0.504 * ( 2371./(E**2.) ) * exp( -5.1909/sqrt(Ecm) ) * exp(-0.548*Ecm) * pEcm


    def _Be7a_pLi6(self, E):
        Q = _eth[16]

        return self._generic_expr(E, Q, 32.6, 10.0, 2.0, 12.0) + self._generic_expr(E, Q, 2.27e6, 8.8335, 13.0, 21.8335)


    def _Be7a_ppnHe4(self, E):
        return self._generic_expr(E, _eth[17], 133.0, 4.0, 3.0, 7.0)

    # END REACTIONS #################################################


    @_convert_mb_to_iMeV2
    def get_cross_section(self, reaction_id, E):
        # There is no switch statement in python :(
        if reaction_id ==  1:
            return self._da_np(E)              #  1. d + a -> n + p
        if reaction_id ==  2:
            return self._ta_nd(E)              #  2. t + a -> n + d
        if reaction_id ==  3:
            return self._ta_npn(E)             #  3. t + a -> 2n + p
        if reaction_id ==  4:
            return self._He3a_pd(E)            #  4. He3 + a -> p + d
        if reaction_id ==  5:
            return self._He3a_npp(E)           #  5. He3 + a -> n + 2p
        if reaction_id ==  6:
            return self._He4a_pt(E)            #  6. He4 + a -> p + t
        if reaction_id ==  7:
            return self._He4a_nHe3(E)          #  7. He4 + a -> n + He3
        if reaction_id ==  8:
            return self._He4a_dd(E)            #  8. He4 + a -> 2d
        if reaction_id ==  9:
            return self._He4a_npd(E)           #  9. He4 + a -> n + p + d
        if reaction_id == 10:
            return self._Li6a_npHe4(E)         # 10. Li6 + a -> n + p + He4
        if reaction_id == 11:
            return self._Li6a_XA3(E)           # 11. Li7 + a -> X + A3
        if reaction_id == 12:
            return self._Li7a_tHe4(E)          # 12. Li7 + a -> t + He4
        if reaction_id == 13:
            return self._Li7a_nLi6(E)          # 13. Li7 + a -> n + Li6
        if reaction_id == 14:
            return self._Li7a_nnpHe4(E)        # 14. Li7 + a -> 2n + p + He4
        if reaction_id == 15:
            return self._Be7a_He3He4(E)        # 15. Be7 + a -> He3 + He4
        if reaction_id == 16:
            return self._Be7a_pLi6(E)          # 16. Be7 + a -> p + Li6
        if reaction_id == 17:
            return self._Be7a_ppnHe4(E)        # 17. Be7 + a -> 2p + n + He4


        # If no match is found, exit with error
        print_error(
            "Reaction with reaction_id " + str(reaction_id) + " does not exist.",
            "acropolis.nucl.NuclearReactor.get_cross_section"
        )


    def _pdi_rates(self, T):
        EC = me2/(22.*T)
        # Set the maximal energy, serving
        # as a cutoff for the integration
        Emax = min( self._sE0, E_EC_max*EC )
        # For E > me2/T >> EC, the spectrum
        # is strongly suppressed

        # Define a dict containing the rates
        # for the photodisintegration reactions
        # key = reaction_id (from _reactions)
        pdi_rates = {rid:approx_zero for rid in _lrid}

        # Calculate the spectra for the given temperature
        if not flags.universal:
            xsp, ysp = self._sGen.get_spectrum(
                                self._sE0, self._sS0, self._sSC, T
                            )
        else:
            xsp, ysp = self._sGen.get_universal_spectrum(
                                self._sE0, self._sS0, self._sSC, T, offset=5e-2
                            )
            # For performance reasons, also
            # cut the energy at threshold
            Emax = min(self._sE0, EC)

        # Interpolate the photon spectrum (in log-log space)
        # With this procedure it should be sufficient to perform
        # a linear interpolation, which also has less side effects
        Fph = LogInterp(xsp, ysp) # Interpolation on: Emin -> E0
        # Calculate the kernel for the integration in log-space
        def Fph_s(log_E, rid):
            E = exp( log_E )
            
            return Fph( E ) * E * self.get_cross_section(rid, E)

        # Define the total rate of interactions altering the photon spectrum,
        # evaluated at the relevant injection energy E0
        rate_photon_E0 = self._sGen.rate_photon(self._sE0, T)

        # Calculate the different rates by looping over all available reaction_id's
        for rid in _lrid:
            # Calculate the 'delta-term'...
            I_dt = self._sS0[0](T)*self.get_cross_section(rid, self._sE0)/rate_photon_E0
            # ... and use it as an initial value
            pdi_rates[rid] = I_dt # might be zero due to exp. suppression!

            # Only perform the integral for energies above threshold,
            # i.e. do not consider strongly suppressed spectra
            if Emax > _eth[rid]:
                # Perform the integration from the threshold energy to Emax
                with warnings.catch_warnings(record=True) as w:
                    log_Emin, log_Emax = log(_eth[rid]), log(Emax)
                    I_Fs = quad(Fph_s, log_Emin, log_Emax, epsrel=eps, epsabs=0, args=(rid,))

                    if len(w) == 1 and issubclass(w[0].category, IntegrationWarning):
                        print_warning(
                            "Slow convergence when calculating the pdi rates " +
                            "@ rid = %i, T = %.3e, E0 = %.3e, Eth = %.3e" % (rid, T, self._sE0, _eth[rid]),
                            "acropolis.nucl.NuclearReactor._thermal_rates_at"
                        )

                # Add the result of the integral to the 'delta' term
                pdi_rates[rid] += I_Fs[0]

            # Avoid potential zeros
            pdi_rates[rid] = max(approx_zero, pdi_rates[rid])

        # Go home and play
        return pdi_rates


    def get_pdi_grids(self):
        (Tmin, Tmax) = self._sTrg

        NT = int(log10(Tmax/Tmin)*NT_pd)

        # Create an array containing all
        # temperature points ('log spacing')
        Tr = np.logspace( log10(Tmin), log10(Tmax), NT )

        # Create a dictionary to store the pdi
        # rates for all reactions and temperatures
        pdi_grids = {rid:np.zeros(NT) for rid in _lrid}

        start_time = time()
        print_info(
            "Calculating non-thermal spectra and reaction rates.",
            "acropolis.nucl.NuclearReactor.get_pdi_grids",
            verbose_level=1
        )

        # Loop over all the temperatures and
        # calculate the corresponding thermal rates
        for i, Ti in enumerate(Tr):
            progress = 100*i/NT
            print_info(
                "Progress: {:.1f}%".format(progress),
                "acropolis.nucl.NuclearReactor.get_pdi_grids",
                eol="\r", verbose_level=1
            )
            rates_at_i = self._pdi_rates(Ti)
            # Loop over the different reactions
            for rid in _lrid:
                pdi_grids[rid][i] = rates_at_i[rid]

        end_time = time()
        print_info(
            "Finished after {:.1f}s.".format(end_time - start_time),
            "acropolis.nucl.NuclearReactor.get_pdi_grids",
            verbose_level=1
        )

        # Go get some sun
        return (Tr, pdi_grids)


class MatrixGenerator(object):

    def __init__(self, temp, pdi_grids, ii):
        self._sII = ii

        # Save the thermal rates
        self._sTemp     = temp
        self._sPdiGrids = pdi_grids

        # Save the appropriate temperature range
        (self._sTmin, self._sTmax) = temp[0], temp[-1]

        # Interpolate the thermal rates
        self._sPdiIp = self._interp_pdi_grids()


    def _interp_pdi_grids(self):
        # A dict containing all interp. rates; key = reaction_id (from _sReactions)
        interp_grids = {}
        for rid in _lrid:
            # Interpolate the rates between
            # Tmin and Tmax in log-log space
            interp_grids[rid] = LogInterp(
                self._sTemp, self._sPdiGrids[rid], base=10. # fill_value=0.
            )

        return interp_grids


    def _pref_ij(self, state, i, j):
        ris = state[0] # initial state of the reaction
        rfs = state[1] # final state of the reaction

        # Find reactions/decays that have
        # 1. the nucleid 'nr=i' in the final state
        # 2. the nucleid 'nc=j' in the initial state
        if ris == j and rfs[i] != 0:
            return rfs[i]

        # Find reactions/decays that have
        # nc = nr in the initial state
        # (diagonal entries)
        if i == j and ris == j:
            return -1.

        return 0.


    def _pdi_kernel_ij(self, i, j, T):
        matij = 0.

        for rid in _lrid:
            matij += self._pref_ij(_rsig[rid], i, j) * self._sPdiIp[rid](T)

        # Incorporate the time-temperature relation and return
        return matij/( self._sII.dTdt(T) )


    def _dcy_kernel_ij(self, i, j, T):
        matij = 0.

        for did in _ldid:
            matij += self._pref_ij(_dsig[did], i, j) * hbar/_tau[did]

        # Incorporate the time-temperature relation and return
        return matij/( self._sII.dTdt(T) )


    def get_matp(self, T):
        # Generate empty matrices
        mpdi, mdcy = np.zeros( (_nnuc, _nnuc) ), np.zeros( (_nnuc, _nnuc) )

        start_time = time()
        print_info(
            "Calculating final transfer matrix.",
            "acropolis.nucl.MatrixGenerator.get_matp",
            verbose_level=1
        )

        nt = 0
        # Rows: Loop over all relevant nuclei
        for nr in range(_nnuc):
            # Columns: Loop over all relevant nuclei
            for nc in range(_nnuc):
                nt += 1

                progress = 100*nt/_nnuc**2
                print_info(
                    "Progress: {:.1f}%".format(progress),
                    "acropolis.nucl.MatrixGenerator.get_matp",
                    eol="\r", verbose_level=1
                )

                # Define the kernels for the integration in log-log space
                Ik_pdi = lambda y: self._pdi_kernel_ij( nr, nc, exp(y) ) * exp(y)
                Ik_dcy = lambda y: self._dcy_kernel_ij( nr, nc, exp(y) ) * exp(y)

                # Perform the integration (in log-log space)
                mpdi[nr, nc] = quad(Ik_pdi, log(self._sTmax), log(T), epsrel=eps, epsabs=0, limit=100)[0]
                mdcy[nr, nc] = quad(Ik_dcy, log(self._sTmax), log(T), epsrel=eps, epsabs=0, limit=100)[0]

        end_time = time()
        print_info(
            "Finished after {:.1f}ms.".format( 1e3*(end_time - start_time) ),
            "acropolis.nucl.MatrixGenerator.get_matp",
            verbose_level=1
        )

        return (mpdi, mdcy)


    def get_all_matp(self):
        NT = len(self._sTemp)

        start_time = time()
        print_info(
            "Calculating transfer matrices for all temperatures.",
            "acropolis.nucl.MatrixGenerator.get_all_matp",
            verbose_level=2
        )

        all_mpdi = np.zeros( (NT, NY, NY) )
        all_mdcy = np.zeros( (NT, NY, NY) )
        for i, temp in enumerate(self._sTemp):
            progress = 100*i/NT
            print_info(
                "Progress: {:.1f}%".format(progress),
                "acropolis.nucl.MatrixGenerator.get_all_matp",
                eol="\r", verbose_level=2
            )

            all_mpdi[i, :, :], all_mdcy[i, :, :] = self.get_matp(temp)

        end_time = time()
        print_info(
            "Finished after {:.1f}s.".format(end_time - start_time),
            "acropolis.nucl.MatrixGenerator.get_all_matp",
            verbose_level=2
        )

        return self._sTemp, (all_mpdi, all_mdcy)


    def get_final_matp(self):
        return self.get_matp( self._sTmin )



# math
from math import pi, log, log10, exp, sqrt
# numpy
import numpy as np
# scipy
from scipy.integrate import quad, dblquad
# abc
from abc import ABCMeta, abstractmethod

# jit
from acropolis.jit import jit
# db
from acropolis.db import import_data_from_db
from acropolis.db import in_rate_db, interp_rate_db
# cache
from acropolis.cache import cached
# pprint
from acropolis.pprint import print_error
# params
from acropolis.params import me, me2, alpha, re
from acropolis.params import zeta3, pi2
from acropolis.params import Emin, approx_zero, eps, Ephb_T_max
from acropolis.params import NE_pd, NE_min


# _ReactionWrapperScaffold ####################################################

@jit
def _JIT_F(Eph, Ee, Ephb):
    # ATTENTION: Here we use the range given in '10.1103/PhysRev.167.1159',
    # because the translation to 0 < q < 1 is questionable
    if not ( Ephb <= Eph <= 4.*Ephb*Ee*Ee/( me2 + 4.*Ephb*Ee ) ):
        # CHECKED to never happen, since the intergration
        # limits are always chosen appropriately (below)
        return 0.

    G = 4.*Ephb*Ee/me2         # \Gamma_\epsilon
    q = Eph/( G*(Ee - Eph) )   # q

    # ATTENTION:
    # If the last term is (2.-2.*G*q) , Kawasaki
    # If the last term is (2.+2.*G*q) , correct
    return 2.*q*log(q) + (1.+2.*q)*(1.-q) + (G*q)**2. * (1.-q)/(2.+2.*G*q)


@jit
def _JIT_G(Ee, Eph, Ephb):
    # Define the energy of the positron
    Eep = Eph + Ephb - Ee

    # Calculate the valid range for Ee
    # ATTENTION: This range is absent in 'astro-ph/9412055'
    # Here we adopt the original result from
    # 'link.springer.com/content/pdf/10.1007/BF01005624.pdf'
    dE_sqrt  = (Eph - Ephb)*sqrt( 1. - me2/( Eph*Ephb ) )
    Ee_lim_m = ( Eph + Ephb - dE_sqrt )/2.
    Ee_lim_p = ( Eph + Ephb + dE_sqrt )/2.
    # ATTENTION: White et al. impose the range in the soft
    # photon limit, which is more difficult to handle but
    # should lead to the same results, since the pair production
    # kernel ensures that Ephb ~ T << Eph ~ O(MeV)

    if not ( me < Ee_lim_m <= Ee <= Ee_lim_p ):
        # CHECKED to never happen, since the intergration
        # limits are always chosen appropriately (below)
        return 0.

    # Split the function into four summands
    # and calculate all of them separately
    # Ee + Eep = Eph + Ephb
    sud  = 0.
    sud += 4.*( (Ee + Eep)**2. )*log( (4.*Ephb*Ee*Eep)/( me2*(Ee + Eep) ) )/( Ee*Eep )
    sud += ( me2/( Ephb*(Ee + Eep) ) - 1. ) * ( (Ee + Eep)**4. )/( (Ee**2.)*(Eep**2.) )
    # ATTENTION: no additional minus sign in sud[2]
    # It is unclear whether it is a typo or an artifact
    # of scanning the original document
    sud += 2.*( 2.*Ephb*(Ee + Eep) - me2 ) * ( (Ee + Eep)**2. )/( me2*Ee*Eep )
    sud += -8.*Ephb*(Ee + Eep)/me2

    return sud


# _PhotonReactionWrapper ######################################################

@jit
def _JIT_ph_rate_pair_creation_ae(logy, logx, T):
    # Return the integrand for the 2d integral in log-space
    x, y = exp(logx), exp(logy)

    # Define beta as a function of y
    b = sqrt(1. - 4.*me2/y)

    # Define the kernel for the 2d-integral; y = s, x = epsilon_bar
    #                     f/E^2                              s   \sigma_DP
    # ATTENTION: There is an error in 'astro-ph/9412055.pdf'
    # In the integration for \bar{\epsilon}_\gamma the lower
    # limit of integration should be me^2/\epsilon_\gamma
    # (the written limit is unitless, which must be wrong)
    # This limit is a consequence of the constraint on
    # the center-of-mass energy
    sig_pc = .5*pi*(re**2.)*(1.-b**2.)*( (3.-b**4.)*log( (1.+b)/(1.-b) ) - 2.*b*(2.-b**2.) )

    return ( 1./(pi**2) )/( exp(x/T) - 1. ) * y * sig_pc * (x*y)


@jit
def _JIT_ph_kernel_inverse_compton(logx, E, Ep, T):
    # Return the integrand for the 1d-integral in log-space; x = Ephb
    x = exp(logx)

    return _JIT_F(E, Ep, x)*x/( pi2*(exp(x/T) - 1.) ) * x


# _ElectronReactionWrapper ####################################################

@jit
def _JIT_el_rate_inverse_compton(y, x, E, T):
    # Return the integrand for the 2d-integral; y = Eph, x = Ephb
    return _JIT_F(y, E, x)*x/( (pi**2.)*(exp(x/T) - 1.) )


@jit
def _JIT_el_kernel_inverse_compton(logx, E, Ep, T):
    # Define the integrand for the 1d-integral in log-space; x = Ephb
    x = exp(logx)

    return _JIT_F(Ep+x-E, Ep, x)*( x/(pi**2) )/( exp(x/T) - 1. ) * x


@jit
def _JIT_el_kernel_pair_creation_ae(logx, E, Ep, T):
    # Define the integrand for the 1d-integral in log-space; x = Ephb
    x = exp(logx)

    return _JIT_G(E, Ep, x)/( (pi**2.)*(exp(x/T) - 1.) ) * x


@jit
def _JIT_dsdE_Z2(Ee, Eph):
    # Define the energies (here: nucleon is at rest)
    Em = Ee                                                      # E_-
    Ep = Eph - Ee                                                # E_+

    # Define the various parameters that enter the cross-section
    pm = sqrt(Em*Em - me2)                                       # p_-
    pp = sqrt(Ep*Ep - me2)                                       # p_+

    L  = log( (Ep*Em + pp*pm + me2)/(Ep*Em - pp*pm + me2) )      # L

    lm = log( (Em + pm)/(Em - pm) )                              # l_-
    lp = log( (Ep + pp)/(Ep - pp) )                              # l_+

    # Define the prefactor
    pref = alpha*(re**2.)*pp*pm/(Eph**3.)

    # Calculate the infamous 'lengthy expression'
    # Therefore, split the sum into four summands
    sud  = 0.
    sud += -4./3. - 2.*Ep*Em*(pp*pp + pm*pm)/( (pp**2.)*(pm**2.) )
    sud += me2*( lm*Ep/(pm**3.) + lp*Em/(pp**3.) - lp*lm/(pp*pm) )
    sud += L*( -8.*Ep*Em/(3.*pp*pm) + Eph*Eph*((Ep*Em)**2. + (pp*pm)**2. - me2*Ep*Em)/( (pp**3.)*(pm**3.) ) )
    sud += -L*me2*Eph*( lp*(Ep*Em - pp*pp)/(pp**3.) + lm*(Ep*Em - pm*pm)/(pm**3.) )/(2.*pp*pm)

    return pref * sud


# SpectrumGenerator ###########################################################

@jit
def _JIT_set_spectra(F, i, Fi, cond=False):
    F[:, i] = Fi
    # In the strongly compressed regime, manually
    # set the photon spectrum to zero in order to
    # avoid floating-point errors
    if cond:
        F[0, i] = 0.


@jit
def _JIT_solve_cascade_equation(E_grid, G, K, S0, SC, T):
    # Extract the number of particle species...
    NX = len(G)
    # ...and the number of energy points
    NE = len(E_grid)

    dy = log(E_grid[-1]/Emin)/(NE-1)

    # Generate the grid for the different spectra
    # 1. index: X = photon, electron, positron
    # 2. index: Position in the energy grid
    F_grid = np.zeros( (NX, NE) )

    # Calculate F_X(E_0), last index NE-1
    FX_E0 = np.array([
        SC[X,-1]/G[X,-1] + np.sum(K[X,:,-1,-1]*S0[:]/(G[:,-1]*G[X,-1])) for X in range(NX)
    ])
    # -->
    _JIT_set_spectra(F_grid, -1, FX_E0)

    # Loop over all energies
    i = (NE - 1) - 1 # start at the second to last index, NE-2
    while i >= 0: # Counting down
        B = np.zeros( (NX, NX) )
        a = np.zeros( (NX,   ) )

        I = np.identity(NX)
        # Calculate the matrix B and the vector a
        for X in range(NX):
            # Calculate B, : <--> Xp
            B[X,:] = -.5*dy*E_grid[i]*K[X,:,i,i] + G[X,i]*I[X,:]

            # Calculate a
            a[X] = SC[X,i]
            for Xp in range(NX):
                a[X] += K[X,Xp,i,-1]*S0[Xp]/G[Xp,-1] + .5*dy*E_grid[-1]*K[X,Xp,i,-1]*F_grid[Xp,-1]
                for j in range(i+1, NE-1): # Goes from i+1 to NE-2
                    a[X] += dy*E_grid[j]*K[X,Xp,i,j]*F_grid[Xp,j]

        # Solve the system of linear equations of the form BF = a
        _JIT_set_spectra(F_grid, i,
            np.linalg.solve(B, a)
        )

        i -= 1

    # Remove potential zeros
    F_grid = F_grid.reshape( NX*NE )
    for i, f in enumerate(F_grid):
        if f < approx_zero:
            F_grid[i] = approx_zero
    F_grid = F_grid.reshape( (NX, NE) )

    # Define the output array...
    sol = np.zeros( (NX+1, NE) )
    # ...and fill it
    sol[0     , :] = E_grid
    sol[1:NX+1, :] = F_grid

    return sol


###############################################################################


class _ReactionWrapperScaffold(object):

    def __init__(self, ii, db):
        self._sII = ii

        self._sY0     = self._sII.bbn_abundances_0()
        self._sEta    = self._sII.parameter("eta")

        self._sRateDb = db


    # NUMBER DENSITIES of baryons, electrons and nucleons #####################

    def _nb(self, T):
        # gs does not change anymore for the relevant temperature,
        # hence (R0/R)^3 = gs(T)T^3/( gs(T0)T0^3) = (T/T0)^3
        return self._sEta * ( 2.*zeta3/pi2 ) * (T**3.)


    def _ne(self, T):
        # 1: p; 5: He4 (see 'NuclearReactor._nuclei' for all identifiers)
        return ( self._sY0[1] + 2.*self._sY0[5] ) * self._nb(T)


    def _nNZ2(self, T):
        # 1: p; 5: He4 (see 'NuclearReactor._nuclei' for all identifiers)
        return ( self._sY0[1] + 4.*self._sY0[5] ) * self._nb(T)


class _PhotonReactionWrapper(_ReactionWrapperScaffold):

    def __init__(self, ii, db):
        super(_PhotonReactionWrapper, self).__init__(ii, db)


    # RATES ###################################################################
    # E is the energy of the incoming particle
    # T is the temperature of the background photons

    # PHOTON-PHOTON SCATTERING ################################################
    def _rate_photon_photon(self, E, T):
        #if E > me2/T:
        #    return 0.
        expf = exp( -E*T/me2 )

        return 0.151348 * (alpha**4.) * me * (E/me)**3. * (T/me)**6. * expf


    # COMPTON SCATTERING ######################################################
    def _rate_compton(self, E, T):
        x = 2.*E/me

        return ( 2.*pi*(re**2.)/x ) * self._ne(T) * ( (1. - 4./x - 8./(x**2.))*log(1.+x) + .5 + 8./x - 1./(2.*(1.+x)**2.) )


    # BETHE-HEITLER PAIR CREATION #############################################
    def _rate_bethe_heitler(self, E, T):
        # For small energies, the rate can be approximated by a constant
        # (cf. 'hep-ph/0604251') --- NOT USED HERE
        #if E < 4.: E = 4.

        k = E/me

        # Below threshold, the rate vanishes
        # This case never happens since Emin = 1.5 > 2me
        # (see 'acropolis.params')
        if k < 2:
            return 0.

        # Approximation for SMALL energies
        if 2 <= k <= 4:
            r = ( 2.*k - 4. )/( k + 2. + 2.*sqrt(2.*k) )

            return ( alpha**3./me2 ) * self._nNZ2(T) * (2.*pi/3.) * ( (k-2.)/k )**3. * ( \
                     1 + r/2. + (23./40.)*(r**2.) + (11./60.)*(r**3.) + (29./960.)*(r**4.) \
                   )


        # Approximation for LARGE energies
        log2k = log(2.*k)
        # We implement corrections up to order (2./k)**6 ('astro-ph/9412055')
        # This is relevant in order to ensure a smooth transition at k = 4
        return ( alpha**3./me2 ) * self._nNZ2(T) * ( \
                   (28./9.)*log2k - 218./27. \
                 + (2./k)**2. * ( (2./3.)*log2k**3. - log2k**2. + (6. - pi2/3.)*log2k + 2.*zeta3 + pi2/6. - 7./2. ) \
                 - (2./k)**4. * ( (3./16.)*log2k + 1./8. ) \
                 - (2./k)**6. * ( (29./2304.)*log2k - 77./13824. ) \
               )


    # DOUBLE PHOTON TO ELECTRON POSITRON PAIR CREATION ########################
    def _rate_pair_creation_ae(self, E, T):
        # In general, the threshold is E ~ me^2/(22*T)
        # However, here we use a slighlty smaller threshold
        # in order to guarantee a smooth transition
        if E < me2/(50.*T):
            return 0.

        # Define the integration limits from the
        # constraint on the center-of-mass energy
        llim = me2/E            # <  50*T (see above)
        ulim = Ephb_T_max*T     # ~ 200*T
        # ulim > llim, since me2/E < 50*T
        # CHECKED!

        # Perform the integration in log-log space
        # The limits for s are always in ascending order,
        # i.e. 4*me2 < 4*E*x, since x > me2/E
        I_fso_E2 = dblquad(_JIT_ph_rate_pair_creation_ae, log(llim), log(ulim), \
                             lambda logx: log(4.*me2), lambda logx: log(4.*E) + logx, \
                             epsrel=eps, epsabs=0, args=(T,)
                          )

        return I_fso_E2[0]/( 8.*E**2. )


    def _rate_pair_creation_ae_db(self, E, T):
        if E < me2/(50.*T):
            return 0.

        E_log, T_log = log10(E), log10(T)
        if ( self._sRateDb is None ) or ( not in_rate_db(E_log, T_log) ):
            return self._rate_pair_creation_ae(E, T)

        return interp_rate_db(self._sRateDb, 'ph:rate_pair_creation_ae', E_log, T_log)


    # TOTAL RATE ##############################################################
    def total_rate(self, E, T):
        return self._rate_photon_photon(E, T) + self._rate_compton(E, T) + self._rate_bethe_heitler(E, T) + self._rate_pair_creation_ae_db(E, T)


    # INTEGRAL KERNELS ########################################################
    # E  is the energy of the outgoing particle
    # Ep is the energy of the incoming particle
    # T  is the temperature of the background photons

    # PHOTON-PHOTON SCATTERING ################################################
    def _kernel_photon_photon(self, E, Ep, T):
        #if Ep > me2/T:
        #    return 0.
        expf = exp( -Ep*T/me2 )

        return 1112./(10125.*pi) * (alpha**4.)/(me**8.) * 8.*(pi**4.)*(T**6.)/63. \
               * Ep**2. * ( 1. - E/Ep + (E/Ep)**2. )**2. * expf


    # COMPTON SCATTERING ######################################################
    def _kernel_compton(self, E, Ep, T):
        # Check that the energies do not execeed the 'Compton edge'
        # ATTENTION: This constraint is missing in '1503.04852'
        if Ep/(1. + 2.*Ep/me) > E:
            return 0.

        # ATTENTION:
        # If the last term is + 2.*me*(1./E - 1./Ep) , Serpico
        # If the last term is - 2.*me*(1./E - 1./Ep) , correct
        return pi*(re**2.)*me/(Ep**2.) * self._ne(T) * ( Ep/E + E/Ep + (me/E - me/Ep)**2. - 2.*me*(1./E - 1./Ep) )


    # INVERSE COMPTON SCATTERING ##############################################
    @cached
    def _kernel_inverse_compton(self, E, Ep, T):
        # Incorporate the non-generic integration limit as
        # the algorithm requires Ep > E and not Ep > E + me
        if Ep < E + me:
            return 0.
        # This also ensures that Ep != E (!!!)

        # Define the integration limits from
        # the range that is specified in '_JIT_F'
        llim = .25*me2*E/( Ep*Ep - Ep*E ) # with Ep != E (see above)
        ulim = min( E, Ep-me2/(4.*Ep), Ephb_T_max*T )
        # Here, the second condition is redundant, since
        # Ep-me2/(4.*Ep) > E for Ep > E + me (see above)
        # However, we include it anyways in order to
        # provide a better documentation
        # CHECKED!

        # If the lower limit exceeds the upper limit,
        # simply return 0. This also helps to avoid
        # overflow if llim > Ephb_Tmax*T
        if ulim <= llim:
            return 0.

        # Perform the integration in log space
        I_fF_E = quad(_JIT_ph_kernel_inverse_compton, log(llim), log(ulim), epsrel=eps, epsabs=0, args=(E, Ep, T))

        # ATTENTION: Kawasaki considers a combined e^+/e^- spectrum
        # Therefore the factor 2 should not be there in our case
        return 2.*pi*(alpha**2.)*I_fF_E[0]/(Ep**2.)


    # TOTAL INTEGRAL KERNEL ####################################################
    def total_kernel_x(self, E, Ep, T, X):
        if X == 0:
            return self._kernel_photon_photon(E, Ep, T) + self._kernel_compton(E, Ep, T)
        # Photon -> Photon

        if X == 1:
            return self._kernel_inverse_compton(E, Ep, T)
        # Electron -> Photon

        if X == 2:
            return self._kernel_inverse_compton(E, Ep, T)
        # Positron -> Photon

        print_error(
            "Particle with identifier X =" + str(X) + "does not exist.",
            "acropolis.cascade._PhotonReactionWrapper.total_kernel_x"
        )


class _AbstractElectronReactionWrapper(_ReactionWrapperScaffold, metaclass=ABCMeta):

    def __init__(self, ii, db):
        super(_AbstractElectronReactionWrapper, self).__init__(ii, db)


    # RATES ###################################################################
    # E is the energy of the incoming particle
    # T is the temperature of the background photons

    # INVERSE COMPTON SCATTERING ##############################################
    @cached
    def _rate_inverse_compton(self, E, T):
        # Define the upper limit for the integration over x
        ulim = min( E - me2/(4.*E), Ephb_T_max*T )
        # The condition x <= E-me2/(4.*E) ensures
        # E- <= E+ in E- <= E <= E+ (range for y)
        # CHECKED!

        # Perform the two-dimensional integration
        # with limits that are calculated from the
        # range that is specified in '_JIT_F'
        # ATTENTION:
        # The integral over \epsilon_\gamma should start at 0.
        # In fact, for \epsilon_\gamma > \epsilon_e, we have q < 0.
        I_fF_E = dblquad(_JIT_el_rate_inverse_compton, 0., ulim, lambda x: x, lambda x: 4.*x*E*E/( me2 + 4.*x*E ), epsrel=eps, epsabs=0, args=(E, T))

        return 2.*pi*(alpha**2.)*I_fF_E[0]/(E**2.)


    def _rate_inverse_compton_db(self, E, T):
        E_log, T_log = log10(E), log10(T)
        if ( self._sRateDb is None ) or ( not in_rate_db(E_log, T_log) ):
            return self._rate_inverse_compton(E, T)

        return interp_rate_db(self._sRateDb, 'el:rate_inverse_compton', E_log, T_log)


    # TOTAL RATE ##############################################################
    def total_rate(self, E, T):
        return self._rate_inverse_compton_db(E, T)


    # INTEGRAL KERNELS ########################################################
    # E  is the energy of the outgoing particle
    # Ep is the energy of the incoming particle
    # T  is the temperature of the background photons

    # INVERSE COMPTON SCATTERING ##############################################
    @cached
    def _kernel_inverse_compton(self, E, Ep, T):
        # E == Ep leads to a divergence in
        # the Bose-Einstein distribution
        # TODO: Check if this can be handled any better
        if E == Ep:
            return 0.

        # Calculate appropriate integration limits
        pf = .25*me2/Ep - E               # <= 0.
        qf = .25*me2*(Ep-E)/Ep            # >= 0.

        sqrt_d = sqrt( (pf/2.)**2. - qf )
        z1 = -pf/2. - sqrt_d              # smaller
        z2 = -pf/2. + sqrt_d              # larger

        # Define the integration limits from
        # the range that is specified in '_JIT_F'
        llim = z1
        ulim = min( z2, Ep - me2/(4.*Ep), Ephb_T_max*T )
        # CHECKED!
        # For the check, remember to use the correct
        # '_JIT_F', i.e. '_JIT_F(Ep+x-E, Ep, x)'

        # If the lower limit exceeds the upper limit,
        # simply return 0. This also helps to avoid
        # overflow if llim > Ephb_Tmax*T
        if ulim <= llim:
            return 0.

        # Perform the integration in log space
        I_fF_E = quad(_JIT_el_kernel_inverse_compton, log(llim), log(ulim), epsrel=eps, epsabs=0, args=(E, Ep, T))

        return 2.*pi*(alpha**2.)*I_fF_E[0]/(Ep**2.)


    # COMPTON SCATTERING ######################################################
    @abstractmethod
    def _kernel_compton(self, E, Ep, T):
        pass


    # BETHE_HEITLER PAIR CREATION #############################################
    @cached
    def _kernel_bethe_heitler(self, E, Ep, T):
        # Incorporate the non-generic integration limit as
        # the algorithm requires Ep > E and not Ep > E + me
        if Ep < E + me:
            return 0.

        # Multiply by the nucleon density and return
        return self._nNZ2(T)*_JIT_dsdE_Z2(E, Ep)


    # DOUBLE PHOTON TO ELECTRON POSITRON PAIR CREATION ########################
    @cached
    def _kernel_pair_creation_ae(self, E, Ep, T):
        # In general, the threshold is Ep >~ me^2/(22*T)
        # However, here we use a slighlty smaller threshold
        # in acordance with the implementation we use in
        # '_PhotonReactionWrapper._rate_pair_creation'
        if Ep < me2/(50.*T):
            return 0.
        # Ep is the incoming(!) energy

        dE, E2 = Ep - E, E**2.
        z1 = Ep*( me2 - 2.*dE*( sqrt(E2 - me2) - E ) )/( 4*Ep*dE + me2 )
        z2 = Ep*( me2 + 2.*dE*( sqrt(E2 - me2) + E ) )/( 4*Ep*dE + me2 )

        # Define the integration limits from
        # the range that is specified in '_JIT_G'
        # and the constraint on the center-of-mass
        # energy, i.e. Eph*Ephb > me^2
        llim = max( me2/Ep, z1 )
        ulim = min( z2, Ep, Ephb_T_max*T )
        # The me < ... condition is fulfiled by
        # default since all energies are larger
        # than Emin > 2me
        # The reference paper also states that
        # x < Ep, which is also incorporated here
        # CHECKED!

        # If the lower limit exceeds the upper limit,
        # simply return 0. This also helps to avoid
        # overflow if llim > Ephb_Tmax*T
        if ulim <= llim:
            return 0.

        # Perform the integration in log space
        I_fG_E2 = quad(_JIT_el_kernel_pair_creation_ae, log(llim), log(ulim), epsrel=eps, epsabs=0, args=(E, Ep, T))

        return 0.25*pi*(alpha**2.)*me2*I_fG_E2[0]/(Ep**3.)


    # TOTAL INTEGRAL KERNEL ####################################################
    @abstractmethod
    def total_kernel_x(self, E, Ep, T, X):
        pass


class _ElectronReactionWrapper(_AbstractElectronReactionWrapper):

    def __init__(self, ii, db):
        super(_ElectronReactionWrapper, self).__init__(ii, db)


    # INTEGRAL KERNELS ########################################################
    # E  is the energy of the outgoing particle
    # Ep is the energy of the incoming particle
    # T  is the temperature of the background photons

    # [...]

    # COMPTON SCATTERING ######################################################
    def _kernel_compton(self, E, Ep, T):
        # Perform a subsitution of the parameters.
        # Compared to the formula for photons, only
        # the arguments of the cross-section are different
        E_s  = Ep + me - E   # E , substituted
        Ep_s = Ep            # Ep, substituted

        # Use the same formula as in case of photons with
        # E  -> E_s
        # Ep -> Ep_s
        # Check that the energies do not exceed the 'Compton edge'
        # ATTENTION: This condition is missing in some other papers
        if Ep_s/(1. + 2.*Ep_s/me) > E_s:
            return 0.

        # ATTENTION:
        # If the last term is + 2.*me*(1./E_s - 1./Ep_s), Serpico
        # If the last term is - 2.*me*(1./E_s - 1./Ep_s), correct
        return pi*(re**2.)*me/(Ep_s**2.) * self._ne(T) * ( Ep_s/E_s + E_s/Ep_s + (me/E_s - me/Ep_s)**2. - 2.*me*(1./E_s - 1./Ep_s) )

    # [...]

    # TOTAL INTEGRAL KERNEL ####################################################
    def total_kernel_x(self, E, Ep, T, X):
        if X == 0:
            return self._kernel_compton(E, Ep, T) + self._kernel_bethe_heitler(E, Ep, T) + self._kernel_pair_creation_ae(E, Ep, T)
        # Photon -> Electron

        if X == 1:
            return self._kernel_inverse_compton(E, Ep, T)
        # Electron -> Electron

        if X == 2:
            return 0.
        # Positron -> Electron

        print_error(
            "Particle with identifier X =" + str(X) + "does not exist.",
            "acropolis.cascade._ElectronReactionWrapper.total_kernel_x"
        )


class _PositronReactionWrapper(_AbstractElectronReactionWrapper):

    def __init__(self, ii, db):
        super(_PositronReactionWrapper, self).__init__(ii, db)


    # INTEGRAL KERNELS ########################################################
    # E  is the energy of the outgoing particle
    # Ep is the energy of the incoming particle
    # T  is the temperature of the background photons

    # [...]

    # COMPTON SCATTERING ######################################################
    def _kernel_compton(self, E, Ep, T):
        # There are (almost) no thermal positrons
        return 0.

    # [...]

    # TOTAL INTEGRAL KERNEL ####################################################
    def total_kernel_x(self, E, Ep, T, X):
        if X == 0:
            return self._kernel_compton(E, Ep, T) + self._kernel_bethe_heitler(E, Ep, T) + self._kernel_pair_creation_ae(E, Ep, T)
        # Photon -> Positron

        if X == 1:
            return 0.
        # Electron -> Positron

        if X == 2:
            return self._kernel_inverse_compton(E, Ep, T)
        # Positron -> Positron

        print_error(
            "Particle with identifier X =" + str(X) + "does not exist.",
            "acropolis.cascade._PositronReactionWrapper.total_kernel_x"
        )


class SpectrumGenerator(object):

    def __init__(self, ii):
        # Extract the data from the databases; If there is
        # no data in the folder 'data/', db = (None, None)
        db = import_data_from_db()

        # Define a dictionary containing all reaction wrappers
        self._sRW = {
            0: _PhotonReactionWrapper  (ii, db),
            1: _ElectronReactionWrapper(ii, db),
            2: _PositronReactionWrapper(ii, db)
        }

         # Set the number of particle species (in the cascade)
        self._sNX = 3


    def _rate_x(self, X, E, T):
        return self._sRW[X].total_rate(E, T)


    def _kernel_x_xp(self, X, Xp, E, Ep, T):
        return self._sRW[X].total_kernel_x(E, Ep, T, Xp)


    def rate_photon(self, E, T):
        return self._rate_x(0, E, T)


    def get_spectrum(self, E0, S0f, SCf, T, allX=False):
        # Define the dimension of the grid
        # from the params in 'params.py'...
        NE = int(log10(E0/Emin)*NE_pd)
        # ... but do not use less than NE_min
        # points
        NE = max(NE, NE_min)

        # Save the dimension of the species grid
        NX = self._sNX

        # Generate the grid for the energy
        E_grid = np.logspace(log(Emin), log(E0), NE, base=np.e)

        # Generate the grid for the different species
        X_grid = np.arange(NX)

        # Generate the grid for the rates
        G = np.array([[self._rate_x(X, E, T) for E in E_grid] for X in X_grid])
            # first index: X, second index according to energy E

        # Generate the grid for the kernels
        K = np.array([[[[self._kernel_x_xp(X, Xp, E, Ep, T) if Ep >= E else 0. for Ep in E_grid] for E in E_grid] for Xp in X_grid] for X in X_grid])
            # first index: X, second index: Xp
            # third index according to energy E
            # fourth index according to energy Ep;
            # For Ep < E, the kernel is simply 0.

        # Generate the grids for the source terms
        # monochromatic + continuous
        S0 = np.array([ S0X(T)                     for S0X in S0f])
        SC = np.array([[SCX(E, T) for E in E_grid] for SCX in SCf])

        # Calculate the spectra by solving
        # the cascade equation
        sol = _JIT_solve_cascade_equation(E_grid, G, K, S0, SC, T)

        # 'sol' always has at least two columns
        return sol[0:2,:] if not allX else sol


    def get_universal_spectrum(self, E0, S0f, SCf, T, offset=0.):
        # Define EC and EX as in 'astro-ph/0211258'
        EC = me2/(22.*T)
        EX = me2/(80.*T)

        # Define the normalization K0 as in 'astro-ph/0211258'
        K0 = E0/( (EX**2.) * ( 2. + log( EC/EX ) ) )

        # Define the dimension of the grid
        # as defined in 'params.py'...
        NE = int(log10(E0/Emin)*NE_pd)
        # ... but not less than NE_min points
        NE = max(NE, NE_min)

        # Generate the grid for the energy
        E_grid = np.logspace(log(Emin), log(E0), NE, base=np.e)
        # Generate the grid for the photon spectrum
        F_grid = np.zeros(NE)

        # Calculate the spectrum for the different energies
        # TODO: Incoporate the continuous source terms in the
        #       normalization by integrating it over the energy
        SN = lambda T: sum(S0X(T) for S0X in S0f) # Normalization
        for i, E in enumerate(E_grid):
            if E < EX:
                F_grid[i] = SN(T) * K0 * (EX/E)**1.5/self.rate_photon(E, T)
            elif E >= EX and E <= (1. + offset)*EC: # an offset enables better interpolation
                F_grid[i] = SN(T) * K0 * (EX/E)**2.0/self.rate_photon(E, T)

        # Remove potential zeros
        F_grid[F_grid < approx_zero] = approx_zero

        # Define the output array...
        sol = np.zeros( (2, NE) )
        # ...and fill it
        sol[0, :] = E_grid
        sol[1, :] = F_grid

        return sol



# functools
from functools import partial
# numpy
import numpy as np
# time
from time import time, sleep
# multiprocessing
from multiprocessing import Pool, cpu_count

# pprint
from acropolis.pprint import print_info, print_error
# models
from acropolis.models import AbstractModel


class ScanParameter(object):

    def __init__(self, ivalue, fvalue, num, spacing="log", fast=False):
        self._sInitialValue  = ivalue
        self._sFinalValue    = fvalue
        self._sSpacingMode   = spacing
        self._sFastParameter = fast
        self._sNumPoints     = num


    def get_range(self):
        if   self._sSpacingMode == "log":
            return np.logspace( self._sInitialValue, self._sFinalValue, self._sNumPoints )
        elif self._sSpacingMode == "lin":
            return np.linspace( self._sInitialValue, self._sFinalValue, self._sNumPoints )


    def is_fast(self):
        return self._sFastParameter


class BufferedScanner(object):

    def __init__(self, model, **kwargs):
        self._sModelClass = model.func if type(model) is partial else model
        
        # Store the requested model
        # self._sModel(...) afterwards creates
        # a new instance of the requested model
        if not issubclass(self._sModelClass, AbstractModel):
            print_error(
                self._sModelClass.__name__ + " is not a subclass of AbstractModel",
                "acropolis.scans.BufferedScanner.__init__"
            )

        self._sModel = model
        
        #######################################################################

        # Initialize the sets that store the verious parameters
        self._sFixed = {}  # Fixed parameters
        self._sScanp = {}  # Scan parameters

        # Initialize the number of scan parameters
        self._sNP      = 0 # (all)
        self._sNP_fast = 0 # (only fast)

        # Initialize the key of the 'fast' parameter
        self._sFast_id = None

        # Parse the keyword arguments and build up the
        # sets 'self._sFixed' and 'self._sScanp'
        self._parse_arguments(**kwargs)

        #######################################################################

        # Extract the keys of the scan parameters
        self._sScanp_id = list( self._sScanp.keys() )


    def _parse_arguments(self, **kwargs):
        # Loop over the different parameters
        for key in kwargs.keys():
            param = kwargs[key]

            # Extract the scan parameters
            if isinstance(param, ScanParameter):
                self._sNP += 1

                # Save the relevant parameter range
                self._sScanp[key] = param.get_range()

                # Check if the parameter is 'fast'
                if param.is_fast():
                    self._sNP_fast += 1

                    # Save the corresponding id
                    self._sFast_id = key
            # Extract the fixed parameters
            else:
                self._sFixed[key] = param

        # Exit in case of unsupported scenarios
        if self._sNP_fast > 1 or self._sNP != 2:
            print_error(
                "The class BufferedScanner only supports 2d parameter scans with <= 1 fast parameters!",
                "acropolis.scans.BufferedScanner._parse_arguments"
            )


    def _rescale_matp_buffer(self, buffer, factor):
        return (factor*buffer[0], buffer[1])


    def _run_single(self, scanp_set):
        assert self._sFast_id is None and self._sNP_fast == 0 and self._sNP == 2
        
        # Define the set that contains all parameters
        fullp_set = scanp_set.copy()
        fullp_set.update( self._sFixed )

        ##############################################################
        Yb = self._sModel(**fullp_set).run_disintegration()
        ##############################################################

        # Sort the scan parameters to guarantee a consistent output
        sorted_scanp = [scanp_set[key] for key in sorted(scanp_set)]

        return [*sorted_scanp, *Yb.transpose().reshape(Yb.size)]


    def _run_batch(self, pp_set):
        assert self._sFast_id is not None and self._sNP_fast == 1 and self._sNP == 2

        # Extract the range of 'fast' parameters to loop over
        fast_params = self._sScanp[self._sFast_id]

        # Initialize the results array
        results = []

        # Initialize the buffer
        matpb = None

        # Initialize fast_param_0
        fast_param_0 = fast_params[0]

        for fast_param in fast_params:
            # Define the set that contains all parameters
            scanp_set = pp_set.copy()
            scanp_set.update( {self._sFast_id: fast_param} )
            # -->
            fullp_set = scanp_set.copy()
            fullp_set.update( self._sFixed )

            # Initialize the model of choice
            model = self._sModel(**fullp_set)

            if matpb is not None:
                # matpb might still be None if E0 < Emin

                factor = fast_param/fast_param_0
                model.set_matp_buffer( self._rescale_matp_buffer(matpb, factor) )

            ##############################################################
            Yb = model.run_disintegration()
            ##############################################################

            # Set up the rescaling
            if matpb is None:
                matpb = model.get_matp_buffer()

                fast_param_0 = fast_param
            
            # Sort the scan parameters to guarantee a consistent output
            sorted_scanp = [scanp_set[key] for key in sorted(scanp_set)]
            # -->
            results.append(
                [*sorted_scanp, *Yb.transpose().reshape(Yb.size)]
            )
        
        return results


    def perform_scan(self, cores=1):
        assert self._sNP_fast <= 1 and self._sNP == 2

        num_cpus = cpu_count() if cores == -1 else cores

        start_time = time()
        print_info(
            "Running scan for {} on {} cores.".format(self._sModelClass.__name__, num_cpus),
            "acropolis.scans.BufferedScanner.perform_scan",
            verbose_level=3
        )

        key1, key2 = self._sScanp_id
        # Generate all possible parameter combinations
        if self._sNP_fast == 0:
            map_params = [
                {key1: val1, key2: val2} for val1 in self._sScanp[key1] \
                                         for val2 in self._sScanp[key2]
            ]

            map_func = self._run_single
        else: # self._sNP_fast == 1
            # Extract the non-fast, i.e. parallel, parameter
            keyp = key1 if key1 != self._sFast_id else key2

            map_params = [
                {keyp: valp} for valp in self._sScanp[keyp]
            ]

            map_func = self._run_batch

        with Pool(processes=num_cpus) as pool:
            # Loop over all possible parameter combinations
            async_results = pool.map_async(map_func, map_params, 1)

            progress = 0
            # Track and print the progress
            while ( progress < 100 ) or ( not async_results.ready() ):
                progress = 100*( len(map_params) - async_results._number_left )/len(map_params)
                print_info(
                    "Progress: {:.1f}%".format(progress),
                    "acropolis.scans.BufferedScanner.perform_scan",
                    eol="\r", verbose_level=3
                )

                sleep(1)

            parallel_results = async_results.get()
            pool.terminate()

        end_time = time()
        print_info(
            "Finished after {:.1f}min.".format( (end_time - start_time)/60 ),
            "acropolis.scans.BufferedScanner.perform_scan",
            verbose_level=3
        )

        parallel_results = np.array(parallel_results)

        # Reshape the array and return
        old_shape = parallel_results.shape
        # -->
        new_shape = (
            np.prod(old_shape[:-1]), old_shape[-1]
        )

        return parallel_results.reshape(new_shape)



# math
from math import log10, floor, ceil
# numpy
import numpy as np
# matplotlib
import matplotlib.pyplot as plt
from matplotlib.ticker import FixedLocator, FixedFormatter
# warnings
import warnings

# obs
from acropolis.obs import pdg2022
# pprint
from acropolis.pprint import print_info
# params
from acropolis.params import NY


# Set the general style of the plot
plt.rc('text', usetex=True)
plt.rc('font', family='serif', size=14)

# Include additional latex packages
plt.rcParams['text.latex.preamble'] = r'\usepackage{amsmath}\usepackage{mathpazo}'


# A global variable counting the
# number of created plots, in order
# to provide unique plot identifiers
_plot_number = 0


# The number of sigmas beyond which a
# point is considered excluded
_95cl = 1.95996 # 95% C.L.


# DATA EXTRACTION ###################################################

def _get_abundance(data, i):
    cols = data.shape[1]

    offset = cols - 3*NY
    # Add + 2 for the two parameters in the first two columns
    i0 = i + offset

    # Extract the different abundances...
    mean, high, low = data[:,i0], data[:,i0+NY], data[:,i0+2*NY]
    # ...and calculate an estimate for the error
    diff = np.minimum( np.abs( mean - high ), np.abs( mean - low ) )

    return mean, diff


def _get_deviations(data, obs):
    # Extract and sum up neutrons and protons
    mn, en     = _get_abundance(data, 0)
    mp, ep     = _get_abundance(data, 1)
    mH, eH     = mn + mp, np.sqrt( en**2. + ep**2. )

    # Extract and sum up lithium-7 and berylium-7
    mLi7, eLi7 = _get_abundance(data, 7)
    mBe7, eBe7 = _get_abundance(data, 8)
    m7,   e7   = mLi7 + mBe7, np.sqrt( eLi7**2. + eBe7**2. )

    # Extract deuterium
    mD  , eD   = _get_abundance(data, 2)

    # Extract and sum up tritium and helium-3
    mT  , eT   = _get_abundance(data, 3)
    mHe3, eHe3 = _get_abundance(data, 4)
    m3  , e3   = mT + mHe3, np.sqrt( eT**2. + eHe3**2. )

    # Extract helium-4
    mHe4, eHe4 = _get_abundance(data, 5)

    # Calculate the actual deviations
    with warnings.catch_warnings(record=True) as w:
        # Calculate the relevant abundance ratios
        mYp , eYp  = 4.*mHe4, 4.*eHe4
        mDH , eDH  = mD/mH, (mD/mH)*np.sqrt( (eD/mD)**2. + (eH/mH)**2. )
        mHeD, eHeD = m3/mD, (m3/mD)*np.sqrt( (e3/m3)**2. + (eD/mD)**2. )
        mLiH, eLiH = m7/mH, (m7/mH)*np.sqrt( (e7/m7)**2. + (eH/mH)**2. )

        # Calculate the corresponding deviations
        Yp  = (mYp  - obs[ 'Yp'].mean) / np.sqrt( obs[ 'Yp'].err**2. +  eYp**2. )
        DH  = (mDH  - obs[ 'DH'].mean) / np.sqrt( obs[ 'DH'].err**2. +  eDH**2. )
        HeD = (mHeD - obs['HeD'].mean) / np.sqrt( obs['HeD'].err**2. + eHeD**2. )
        LiH = (mLiH - obs['LiH'].mean) / np.sqrt( obs['LiH'].err**2. + eLiH**2. )

        if len(w) == 1 and issubclass(w[0].category, RuntimeWarning):
            # Nothing to do here
            pass

    # Take care of potential NaNs
    HeD[ mDH  < obs['DH' ].err ] =  10
    HeD[ mHeD < obs['HeD'].err ] =  10
    DH [     np.isnan(DH)      ] = -10

    # Return (without reshaping)
    return Yp, DH, HeD, LiH


# LATEX INFORMATION #################################################


_tex_data = {
    # DecayModel
    'mphi'  : (r'm_\phi'                                         , r'\mathrm{MeV}'   ),
    'tau'   : (r'\tau_\phi'                                      , r'\mathrm{s}'     ),
    'temp0' : (r'T_0'                                            , r'\mathrm{MeV}'   ),
    'n0a'   : (r'(n_\phi/n_\gamma)|_{T=T_0}'                     , r''               ),
    # AnnihilationModel
    'braa'  : (r'\text{BR}_{\gamma\gamma} = 1-\text{BR}_{e^+e^-}', r''               ),
    'mchi'  : (r'm_\chi'                                         , r'\mathrm{MeV}'   ),
    'a'     : (r'a'                                              , r'\mathrm{cm^3/s}'),
    'b'     : (r'b'                                              , r'\mathrm{cm^3/s}'),
    'tempkd': (r'T_\text{kd}'                                    , r'\mathrm{MeV}'   ),
}


def add_tex_data(key, tex, unit):
    global _tex_data

    _tex_data[key] = (tex, unit)


def tex_title(**kwargs):
    global _tex_data

    if _tex_data is None:
        return

    eof = r',\;'

    # Define a function to handle values
    # that need to be printed in scientific
    # notation
    def _val_to_string(val):
        if isinstance(val, float):
            power = log10( val )
            if power != int(power):
                # TODO: Implement this special case
                pass

            return r'10^' + str( int(power) )

        return str( val )

    title = r''
    for key in kwargs.keys():
        # Extract the numerical value
        val     = kwargs[ key ]
        val_str = _val_to_string( val )
        # Extract the latex representation
        # of the parameter and its unit
        tex, unit = _tex_data[ key ]
        # If the value is 0, do not print units
        unit = '\,' + unit if val != 0 else r''

        title += tex + '=' + val_str + unit + eof

    if title.endswith(eof):
        title = title[:-len(eof)]

    return r'$' + title + r'$'


def tex_label(key):
    if key not in _tex_data.keys():
        return ''

    tex, unit = _tex_data[ key ]

    if unit != r'':
        unit = r'\;[' + unit + r']'

    return r'$' + tex + unit + r'$'


def tex_labels(key_x, key_y):
    return ( tex_label(key_x), tex_label(key_y) )


# FIGURE HANDLING ###################################################

def init_figure():
    fig = plt.figure(figsize=(4.8, 4.4), dpi=150, edgecolor='white')
    ax  = fig.add_subplot(1, 1, 1)

    ax.tick_params(axis='both', which='both', labelsize=11, direction='in', width=0.5)

    ax.xaxis.set_ticks_position('both')
    ax.yaxis.set_ticks_position('both')
    for axis in ['top','bottom','left','right']:
        ax.spines[axis].set_linewidth(0.5)

    return fig, ax


def _set_tick_labels(ax, x, y):
    nint = lambda val: ceil(val) if val >= 0 else floor(val)

    xmin, xmax = np.min(x), np.max(x)
    ymin, ymax = np.min(y), np.max(y)

    xmin_log = nint( log10(xmin) )
    xmax_log = nint( log10(xmax) )
    ymin_log = nint( log10(ymin) )
    ymax_log = nint( log10(ymax) )

    nx = abs( xmax_log - xmin_log ) + 1
    ny = abs( ymax_log - ymin_log ) + 1

    # Set the ticks on the x-axis
    xticks_major  = np.linspace(xmin_log, xmax_log, nx)
    xticks_minor  = [ log10(i*10**j) for i in range(1, 10) for j in xticks_major ]
    xlabels       = [ r'$10^{' + f'{int(i)}' + '}$' for i in xticks_major ]

    xticks_major_locator = FixedLocator(xticks_major)
    xticks_minor_locator = FixedLocator(xticks_minor)
    xlabels_formatter    = FixedFormatter(xlabels)

    ax.xaxis.set_major_locator(xticks_major_locator)
    ax.xaxis.set_minor_locator(xticks_minor_locator)
    ax.xaxis.set_major_formatter(xlabels_formatter)
    ax.set_xlim(xmin_log, xmax_log)

    # Set the ticks on the y-axis
    yticks_major  = np.linspace(ymin_log, ymax_log, ny)
    yticks_minor  = [ log10(i*10**j) for i in range(1, 10) for j in yticks_major ]
    ylabels       = [ r'$10^{' + f'{int(i)}' + '}$' for i in yticks_major ]

    yticks_major_locator = FixedLocator(yticks_major)
    yticks_minor_locator = FixedLocator(yticks_minor)
    ylabels_formatter    = FixedFormatter(ylabels)

    ax.yaxis.set_major_locator(yticks_major_locator)
    ax.yaxis.set_minor_locator(yticks_minor_locator)
    ax.yaxis.set_major_formatter(ylabels_formatter)
    ax.set_ylim(ymin_log, ymax_log)


def save_figure(output_file=None, show_fig=False):
    global _plot_number

    # If no name for the output file is given
    # simply enumerate the different plots
    if output_file is None:
        output_file = 'acropolis_plot_{}.pdf'.format(_plot_number)
        _plot_number += 1

    plt.savefig(output_file)
    # Show the figure on request
    if show_fig:
        plt.show()

    print_info(
        "Figure has been saved as '{}'".format(output_file),
        "acropolis.plot.save_figure"
    )


def plot_scan_results(data, output_file=None, contour_file=None, title='', labels=('', ''), fix_helium=False, show_fig=False, obs=pdg2022, xc=0, yc=1):
    # If data is a filename, load the data first
    if isinstance(data, str):
        data = np.loadtxt(data)

    # Get the set of input parameters...
    x, y = data[:,xc], data[:,yc]

    # ...and determine the shape of the data
    N  = len(x)
    Ny = (x == x[0]).sum()
    Nx = N//Ny

    shape = (Nx, Ny)

    # Calculate the abundance deviations
    Yp, DH, HeD, LiH = _get_deviations(data, obs)

    # Reshape the input data...
    x   =   x.reshape(shape)
    y   =   y.reshape(shape)
    # ...and the deviation arrays
    Yp  =  Yp.reshape(shape)
    DH  =  DH.reshape(shape)
    HeD = HeD.reshape(shape)
    LiH = LiH.reshape(shape)

    # Fix potential 'holes' in the
    # exclusion region of HeD
    if fix_helium:
        for j, column in enumerate(HeD.T):

            excl = False
            for i, el in enumerate(column):
                if el < _95cl and excl:
                    HeD[i, j] = 10

                if el > _95cl and not excl:
                    excl = True

    # Extract the overall exclusion limit
    max = np.maximum( np.abs(DH), np.abs(Yp) )
    max = np.maximum( max, HeD )

    # Init the figure and...
    fig, ax = init_figure()
    # ...set the tick labels
    _set_tick_labels(ax, x, y)

    # Plot the actual data
    cut = 1e10
    # Deuterium (filled)
    ax.contourf(np.log10(x), np.log10(y), DH,
        levels=[-cut, -_95cl, _95cl, cut],
        colors=['0.6','white', 'tomato'],
        alpha=0.2
    )
    # Helium-4 (filled)
    ax.contourf(np.log10(x), np.log10(y), Yp,
        levels=[-cut, -_95cl, _95cl, cut],
        colors=['dodgerblue','white', 'lightcoral'],
        alpha=0.2
    )
    # Helium-3 (filled)
    ax.contourf(np.log10(x), np.log10(y), HeD,
        levels=[_95cl, cut], # Only use He3/D as an upper limit
        colors=['mediumseagreen'],
        alpha=0.2
    )

    # Deuterium low (line)
    ax.contour(np.log10(x), np.log10(y), DH,
        levels=[-_95cl], colors='0.6', linestyles='-'
    )
    # Deuterium high (line)
    ax.contour(np.log10(x), np.log10(y), DH,
        levels=[_95cl], colors='tomato', linestyles='-'
    )
    # Helium-4 low (line)
    ax.contour(np.log10(x), np.log10(y), Yp,
        levels=[-_95cl], colors='dodgerblue', linestyles='-'
    )
    # Helium-3 high (line)
    ax.contour(np.log10(x), np.log10(y), HeD,
        levels=[_95cl], colors='mediumseagreen', linestyles='-'
    )
    # Overall high/low (line)
    cs = ax.contour(np.log10(x), np.log10(y), max,
        levels=[_95cl], colors='black', linestyles='-'
    )

    # Set the title...
    ax.set_title( title, fontsize=11 )
    # ...and the axis labels
    ax.set_xlabel( labels[0] )
    ax.set_ylabel( labels[1] )

    # Set tight layout
    plt.tight_layout()

    if contour_file is not None:
        extraction_failed = False

        if len(cs.collections) == 1:
            paths = cs.collections[0].get_paths()

            if len(paths) == 1:
                xy_cs = paths[0].vertices

                np.savetxt(contour_file, xy_cs)

                print_info(
                    "Overall exclusion line has been saved as '{}'".format(contour_file),
                    "acropolis.plot.plot_scan_results"
                )
            else:
                extraction_failed = True
        else:
            extraction_failed = True
        
        if extraction_failed:
            print_info(
                "Could not extract unique contour. No data has been saved.",
                "acropolis.plot.plot_scan_results"
            )

    if output_file is not None:
        save_figure(output_file)

    if show_fig:
        plt.show()

    # Return figure and axis in case
    # further manipulation is desired
    return fig, ax



# math
from math import log, pow
# numpy
import numpy as np
# scipy
from scipy.integrate import cumulative_simpson


class LogInterp(object):

    def __init__(self, x_grid, y_grid, base=np.e, fill_value=None):
        self._sBase    = base
        self._sLogBase = log(self._sBase)

        self._sFillValue = fill_value

        self._sXLog = np.log(x_grid)/self._sLogBase
        self._sYLog = np.log(y_grid)/self._sLogBase

        self._sXminLog = self._sXLog[ 0]
        self._sXmaxLog = self._sXLog[-1]

        xdiff = np.diff(self._sXLog)
        if not np.all( xdiff >= 0 ):
            raise ValueError(
                "The values in x_grid need to be in ascending order."
            )
        if not np.allclose( xdiff, xdiff[0] ):
            raise ValueError(
                "The values in x_grid need to be equidistant in log space."
            )

        self._sN = len(self._sXLog)

        self._sCache = {}


    def _perform_interp(self, x):
        x_log = log(x)/self._sLogBase

        if not (self._sXminLog <= x_log <= self._sXmaxLog):
            if self._sFillValue is None:
                raise ValueError(
                    "The given value does not lie within the interpolation range."
                )

            return self._sFillValue

        ix = int( ( x_log - self._sXminLog )*( self._sN - 1 )/( self._sXmaxLog - self._sXminLog ) )

        # Handle the case for which ix+1 is out-of-bounds
        if ix == self._sN - 1:
            ix -= 1

        x1_log, x2_log = self._sXLog[ix], self._sXLog[ix+1]
        y1_log, y2_log = self._sYLog[ix], self._sYLog[ix+1]

        m = ( y2_log - y1_log )/( x2_log - x1_log )
        b = y2_log - m*x2_log

        return pow( self._sBase, m*x_log + b )


    def __call__(self, x):
        if x not in self._sCache:
            self._sCache[x] = self._perform_interp(x)

        return self._sCache[x]


# Cummulative numerical Simpson integration
def cumsimp(x_grid, y_grid):
    return cumulative_simpson(x_grid*y_grid, x=np.log(x_grid), initial=0.)



_use_numba_jit = True

# numba
try:
    import numba as nb
except ImportError:
    _use_numba_jit = False


def _null_decorator(f):
    return f


jit = nb.njit(cache=True) if _use_numba_jit else _null_decorator


# functools
from functools import wraps


def cached(f):
    # Define the cache as a dictionary
    cache = {}
    Tc = {"_": -1.}

    # Define the wrapper function
    @wraps(f)
    def f_cached(*args):
        T     = args[-1]
        # Drop the first argument 'self'
        # ! only for member functions !
        pargs = args[1:]

        # For each new temperature,
        # clear the cache and start over
        if T != Tc["_"]:
            Tc["_"] = T
            cache.clear()

        if pargs not in cache:
            cache[pargs] = f(*args)

        return cache[pargs]

    return f_cached



# gzip
import gzip
# pickle
import pickle
# os
from os import path
# time
from time import time

# jit
from acropolis.jit import jit
# pprint
from acropolis.pprint import print_info
# flags
import acropolis.flags as flags
# params
from acropolis.params import Emin_log, Emax_log, Enum
from acropolis.params import Tmin_log, Tmax_log, Tnum


def import_data_from_db():
    pkg_dir, _ = path.split(__file__)
    db_file    = path.join(pkg_dir, "data", "rates.db.gz")

    ratedb = None
    if not flags.usedb or not path.exists(db_file):
        return ratedb

    start_time = time()
    print_info(
        "Extracting and reading database files.",
        "acropolis.db.import_data_from_db",
        verbose_level=1
    )

    ratefl = gzip.open(db_file, "rb")
    ratedb = pickle.load(ratefl)
    ratefl.close()

    end_time = time()
    print_info(
        "Finished after {:.1f}ms.".format( 1e3*(end_time - start_time) ),
        "acropolis.db.import_data_from_db",
        verbose_level=1
    )

    return ratedb


def in_rate_db(E_log, T_log):
    if (Emin_log <= E_log <= Emax_log) and (Tmin_log <= T_log <= Tmax_log):
        return True

    return False


def in_kernel_db(E_log, Ep_log, T_log):
    if (Emin_log <= E_log <= Emax_log) \
      and (Emin_log <= Ep_log <= Emax_log) \
      and (Tmin_log <= T_log <= Tmax_log):
        return True

    return False


@jit
def _get_E_log(i):
    return Emin_log + (Emax_log - Emin_log)*i/(Enum - 1)


@jit
def _get_T_log(i):
    return Tmin_log + (Tmax_log - Tmin_log)*i/(Tnum - 1)


@jit
def _get_E_index(E_log):
    index = int( ( Enum - 1 ) * ( E_log - Emin_log ) / ( Emax_log - Emin_log ) )

    # For points at the upper boundary, i+1 does not exist
    return index if index != Enum - 1 else index - 1


@jit
def _get_T_index(T_log):
    index = int( ( Tnum - 1 ) * ( T_log - Tmin_log ) / ( Tmax_log - Tmin_log ) )

    # For points at the upper boundary, i+1 does not exist
    return index if index != Tnum - 1 else index - 1


@jit
def interp_rate_db(rate_db, id, E_log, T_log):
    # Extract the correct index for the datafile
    c = {
        'ph:rate_pair_creation_ae': 0,
        'el:rate_inverse_compton' : 1
    }[id]

    # Calculate the respective indices in the interpolation file
    iE, iT  = _get_E_index(E_log), _get_T_index(T_log)

    # Perform the interpolation according to the wikipedia page:
    # https://en.wikipedia.org/wiki/Bilinear_interpolation
    x , y = T_log, E_log
    x0, y0 = _get_T_log(iT  ), _get_E_log(iE  )
    x1, y1 = _get_T_log(iT+1), _get_E_log(iE+1)

    # Define the index function
    k = lambda jT, jE: jT*Enum + jE

    c00 = rate_db[ k(iT  , iE  ) ][c]
    c10 = rate_db[ k(iT+1, iE  ) ][c]
    c01 = rate_db[ k(iT  , iE+1) ][c]
    c11 = rate_db[ k(iT+1, iE+1) ][c]

    d = (x0-x1)*(y0-y1)

    a0 = ( c00*x1*y1 - c01*x1*y0 - c10*x0*y1 + c11*x0*y0 )/d
    a1 = ( -c00*y1 + c01*y0 + c10*y1 - c11*y0 )/d
    a2 = ( -c00*x1 + c01*x1 + c10*x0 - c11*x0 )/d
    a3 = ( c00 - c01 - c10 + c11 )/d

    return 10.**( a0 + a1*x + a2*y + a3*x*y )


@jit
def interp_kernel_db(kernel_db, id, E_log, Ep_log, T_log):
    raise NotImplementedError

    """
    # Calculate the respective indices in the interpolation file
    iE, iEp, iT  = _get_E_index(E_log), _get_E_index(Ep_log), _get_T_index(T_log)

    # Perform the interpolation according to the wikipedia page:
    # https://en.wikipedia.org/wiki/Trilinear_interpolation
    x , y , z = T_log, E_log, Ep_log
    x0, y0, z0, = _get_T_log(iT  ), _get_E_log(iE  ), _get_E_log(iEp  )
    x1, y1, z1, = _get_T_log(iT+1), _get_E_log(iE+1), _get_E_log(iEp+1)
    xd, yd, zd = (x-x0)/(x1-x0), (y-y0)/(y1-y0), (z-z0)/(z1-z0)

    # Define the index function
    k = lambda jT, jE, jEp: jT*Enum*(Enum+1)//2 + jE*Enum - (jE-1)*jE//2 + (jEp - jE)

    c000 = kernel_db[ k(iT  , iE  , iEp  ) ][c]
    c100 = kernel_db[ k(iT+1, iE  , iEp  ) ][c]
    c010 = kernel_db[ k(iT  , iE+1, iEp  ) ][c]
    c001 = kernel_db[ k(iT  , iE  , iEp+1) ][c]
    c110 = kernel_db[ k(iT+1, iE+1, iEp  ) ][c]
    c101 = kernel_db[ k(iT+1, iE  , iEp+1) ][c]
    c111 = kernel_db[ k(iT+1, iE+1, iEp+1) ][c]
    c011 = kernel_db[ k(iT  , iE+1, iEp+1) ][c]

    c00 = c000*(1.-xd) + c100*xd
    c01 = c001*(1.-xd) + c101*xd
    c10 = c010*(1.-xd) + c110*xd
    c11 = c011*(1.-xd) + c111*xd

    c0 = c00*(1.-yd) + c10*yd
    c1 = c01*(1.-yd) + c11*yd

    c = c0*(1.-zd) + c1*zd

    d = (x0-x1)*(y0-y1)*(z0-z1)

    a0 = ( -c000*x1*y1*z1 + c001*x1*y1*z0 + c010*x1*y0*z1 - c011*x1*y0*z0 + \
           c100*x0*y1*z1 - c101*x0*y1*z0 - c110*x0*y0*z1 + c111*x0*y0*z0)/d
    a1 = ( c000*y1*z1 - c001*y1*z0 - c010*y0*z1 + c011*y0*z0 + \
           -c100*y1*z1 + c101*y1*z0 + c110*y0*z1 - c111*y0*z0)/d
    a2 = ( c000*x1*z1 - c001*x1*z0 - c010*x1*z1 + c011*x1*z0 + \
           -c100*x0*z1 + c101*x0*z0 + c110*x0*z1 - c111*x0*z0)/d
    a3 = ( c000*x1*y1 - c001*x1*y1 - c010*x1*y0 + c011*x1*y0 + \
           -c100*x0*y1 + c101*x0*y1 + c110*x0*y0 - c111*x0*y0)/d
    a4 = ( -c000*z1 + c001*z0 + c010*z1 - c011*z0 + c100*z1 - c101*z0 - c110*z1 + c111*z0 )/d
    a5 = ( -c000*y1 + c001*y1 + c010*y0 - c011*y0 + c100*y1 - c101*y1 - c110*y0 + c111*y0 )/d
    a6 = ( -c000*x1 + c001*x1 + c010*x1 - c011*x1 + c100*x0 - c101*x0 - c110*x0 + c111*x0 )/d
    a7 = ( c000 - c001 - c010 + c011 - c100 + c101 + c110 - c111 )/d

    return 10.**( a0 + a1*x + a2*y + a3*z + a4*x*y + a5*x*z + a6*y*z + a7*x*y*z )
    """



# os
from os import path
# math
from math import log10
# numpy
import numpy as np
# tarfilfe
import tarfile
# abc
from abc import ABC, abstractmethod

# util
from acropolis.utils import cumsimp
# pprint
from acropolis.pprint import print_error
# params
from acropolis.params import hbar
from acropolis.params import NY, NC


def locate_data_file(filename):
    pkg_dir, _  = path.split(__file__)
    data_file   = path.join(pkg_dir, "data", filename)

    return data_file


def locate_sm_file():
    return locate_data_file("sm.tar.gz")


def input_data_from_dir(dirname):
    cosmo_data = np.genfromtxt(dirname + "/cosmo_file.dat")
    abund_data = np.genfromtxt(dirname + "/abundance_file.dat")
    param_data = np.genfromtxt(dirname + "/param_file.dat", delimiter="=", dtype=None, encoding=None)

    return InputData(cosmo_data, abund_data, param_data)


def input_data_from_file(filename):
    # Read the input file
    tf, tc = tarfile.open(filename, "r:gz"), {}

    # Extract the different files and
    # store them in a dictionary
    for m in tf.getmembers():
        tc[m.name] = tf.extractfile(m)

    # READ THE PREVIOUSLY GENERATED DATA
    cosmo_data = np.genfromtxt(tc["cosmo_file.dat"]    )
    abund_data = np.genfromtxt(tc["abundance_file.dat"])
    param_data = np.genfromtxt(tc["param_file.dat"], delimiter="=", dtype=None, encoding=None)

    return InputData(cosmo_data, abund_data, param_data)


class AbstractData(ABC):

    @abstractmethod
    def get_cosmo_data(self):
        pass

    @abstractmethod
    def get_abund_data(self):
        pass

    @abstractmethod
    def get_param_data(self):
        pass


class InputData(AbstractData):

    def __init__(self, cosmo_data, abund_data, param_data):
        self._sCosmoData = cosmo_data
        self._sAbundData = abund_data
        self._sParamData = param_data


    def get_cosmo_data(self):
        return self._sCosmoData


    def get_abund_data(self):
        return self._sAbundData


    def get_param_data(self):
        return self._sParamData


class InputInterface(object):

    def __init__(self, input_data, type="file"):
        if   type == "file":
            if not isinstance(input_data, str):
                raise ValueError(
                    "If 'type == file', 'input_data' must be a string"
                )

            input_data = input_data_from_file(input_data)      
        elif type == "dir" :
            if not isinstance(input_data, str):
                raise ValueError(
                        "If 'type == dir', 'input_data' must be be a string"
                    )
            
            input_data = input_data_from_dir(input_data)
        elif type == "raw" :
            if not isinstance(input_data, InputData):
                raise ValueError(
                        "If 'type == raw', 'input_data' must be an instance of InputData"
                    )
        else:
            raise ValueError(
                "Unknown type for input data: Only 'file', 'dir' and 'raw' are supported"
            )

        # Extract the provided input data
        self._sCosmoData = input_data.get_cosmo_data()
        self._sAbundData = input_data.get_abund_data()
        self._sParamData = input_data.get_param_data()

        # Calculate the scale factor and add it
        sf = np.exp( cumsimp(self._sCosmoData[:,0]/hbar, self._sCosmoData[:,4]) )
        self._sCosmoData = np.column_stack( [self._sCosmoData, sf] )
        # The corresponding index will be -1,
        # irregardless of the number of additional
        # columns in the cosmo-file

        # Log the cosmo data for the interpolation
        # ATTENTION: At this point we have to take the
        # absolute value, because dT/dt is negative
        self._sCosmoDataLog = np.log10( np.abs(self._sCosmoData) )
        self._sCosmoDataShp = self._sCosmoData.shape

        # Reshape the abundance data
        self._sAbundData = self._sAbundData.reshape(
                                    (NY, self._sAbundData.size//NY)
                                )
        # Reshape the param data and
        # turn it into a dictionary
        self._sParamData = dict( self._sParamData.reshape(self._sParamData.size) )

        # Check if the data is consistent
        self._check_data()


    def _check_data(self):
        # Check if param_file.dat includes the required parameters
        req_param   = ( "eta" in self._sParamData )
        if not req_param:
            print_error(
                "The mandatory variable 'eta' could not be found in 'param_file.dat'",
                "acropolis.input.InputInterface::_check_data"
            )

        # Check if abundance_file.dat has the correct dimensions
        abund_shape = ( self._sAbundData.shape[0] == NY )
        if not abund_shape:
            print_error(
                "The content of 'abundance_file.dat' does not have the required shape.",
                "acropolis.input.InputInterface::_check_data"
            )

        # Check if cosmo_file.dat has the correct number of columns
        cosmo_shape = ( self._sCosmoDataShp[1] >= NC )
        if not cosmo_shape:
            print_error(
                "The content of 'cosmo_file.dat' does not have the required shape.",
                "acropolis.input.InputInterface::_check_data"
            )


    # 1. COSMO_DATA ###########################################################

    def _find_index(self, x, x0):
        # Check if the array is in ascending
        # or decending order
        ascending = x[0] < x[1]
        # -->
        decending = not ascending

        # Calculate the size of the array
        size = self._sCosmoDataShp[0]

        # Perform a binary search
        left  = 0
        right = size - 1

        while left <= right:
            mid = int(left + (right - left) / 2)

            if ((ascending and x[mid] <= x0 and (mid == size - 1 or x[mid + 1] > x0)) or (decending and x[mid] >= x0 and (mid == size - 1 or x[mid + 1] < x0))):
                return mid
            
            if ((ascending and x[mid] < x0) or (decending and x[mid] > x0)):
                left = mid + 1
            else:
                right = mid - 1

        return -1


    def _interp_cosmo_data(self, val, xc, yc):
        # ATTENTION: To ensure maximal performance,
        # it is assumed that x is already sorted in
        # either increasing or decreasing order
        x = self._sCosmoDataLog[:,xc]
        y = self._sCosmoDataLog[:,yc]

        val_log = log10(val)

        # Extract the index ix for which val_log
        # is between x[ix] and x[ix+1]
        ix = self._find_index(x, val_log)

        if ix == -1 or not (x[ix] <= val_log <= x[ix+1] or x[ix] >= val_log >= x[ix+1]):
            raise ValueError(
                "Interpolation error: Index out of range"
            )

        m = (y[ix+1] - y[ix])/(x[ix+1] - x[ix])
        b = y[ix] - m*x[ix]

        return 10.**(m*val_log + b)


    def temperature(self, t):
        return self._interp_cosmo_data(t, 0, 1)


    def time(self, T):
        return self._interp_cosmo_data(T, 1, 0)


    def dTdt(self, T):
        return -self._interp_cosmo_data(T, 1, 2)


    def neutrino_temperature(self, T):
        return self._interp_cosmo_data(T, 1, 3)


    def hubble_rate(self, T):
        return self._interp_cosmo_data(T, 1, 4)


    def scale_factor(self, T):
        return self._interp_cosmo_data(T, 1, -1)


    def cosmo_column(self, yc, val, xc=1):
        return self._interp_cosmo_data(val, xc, yc)


    def temperature_range(self):
        return ( min(self._sCosmoData[:,1]), max(self._sCosmoData[:,1]) )


    # 2. ABUNDANCE_DATA #######################################################

    def bbn_abundances(self):
        return self._sAbundData


    def bbn_abundances_0(self):
        return self._sAbundData[:,0]


    # 3. PARAM_DATA ###########################################################

    def parameter(self, key):
        return self._sParamData[key]


    def param_keys(self):
        return self._sParamData.keys()



# functools
from functools import partial

# params
from acropolis.params import me2
from acropolis.params import pi

# ext.models
from acropolis.ext.models import ResonanceModel, estimate_tempkd_ee


# Benchmark (1)
def sigma_ee_b1(s, mchi, delta, gammad, gammav):
    mx2 = mchi**2.
    mx4 = mchi**4.

    pref = 16. * pi * gammad * gammav / mx4 / (2. + delta)**4. / s**2.
    # -->
    return pref * mx2 * ( s*s - 2.*s*(mx2 - 3.*me2) + (me2 - mx2)**2. )


# Benchmark (2)
def sigma_ee_b2(s, mchi, delta, gammad, gammav):
    mx2 = mchi**2.
    mx4 = mchi**4.

    me4 = me2*me2

    pref = 16. * pi * gammad * gammav / mx4 / (2. + delta)**4. / s**2.
    # -->
    #return 3. * pref * ( 2.*(mx2 - me2)**4. - 8.*(mx4 - me4)*(mx2 - me2)*s \
    #       + (15.*mx4 + 26.*mx2*me2 + 15.*me4)*s*s - 14.*(mx2 + me2)*(s**3.) + 5.*(s**4.) ) / s
    return .25* pref * ( 4.*(s**3.) - 10.*(s**2.)*(mx2 + me2) + s*(9.*mx4 + 22.*mx2*me2 + 9.*me4) - 4.*(mx4 - me4)*(mx2 - me2) + (mx2 - me2)**4./s )


# Benchmark (3)
def sigma_ee_b3(s, mchi, delta, gammad, gammav):
    mx2 = pow(mchi, 2.)
    mx4 = pow(mchi, 4.)

    me4 = me2*me2

    pref = 16. * pi * gammad * gammav / mx4 / (2. + delta)**4. / s**2.
    # -->
    #return 48. * pref * ( mx4 * (me2 + 3.*s) + (me2 - s)**2. * (me2 + 3.*s) - 2.*mx2 * (me2**2. - 4.*me2*s + 3.*s*s) )
    return 4.5 * pref * ( mx4*(me2 + s) - 2.*mx2*(me2 - s)**2. + (me4 - s**2.)*(me2 - s) )


# -->
BenchmarkModel1 = partial(ResonanceModel, nd = 0., S = 1./2., tempkd = partial(estimate_tempkd_ee, sigma_ee=sigma_ee_b1)) # (1)
BenchmarkModel2 = partial(ResonanceModel, nd = 0., S = 3./4., tempkd = partial(estimate_tempkd_ee, sigma_ee=sigma_ee_b2)) # (2)
BenchmarkModel3 = partial(ResonanceModel, nd = 1., S = 3./2., tempkd = partial(estimate_tempkd_ee, sigma_ee=sigma_ee_b3)) # (3)


# setuptools
from setuptools import setup
# os
from os import path

# version
from acropolis.info import version, description, url, authors

# Read the README.md file
cwd = path.abspath( path.dirname(__file__) )
with open(path.join(cwd, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Perform the actual setup
setup(
    name='ACROPOLIS',
    version=version,
    python_requires='>=3.6',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=url,
    author=authors,
    license='GPL3',
    packages=[
        'acropolis',
        'acropolis/ext'
    ],
    package_data={
        'acropolis': ['data/rates.db.gz', 'data/sm.tar.gz', 'data/parthenope_v3_0.tar.gz']
    },
    include_package_data=True,
    scripts=[
        'decay',
        'annihilation'
    ],
    install_requires=[
        'numpy>=1.19.1',
        'scipy>=1.5.2',
        'numba>=0.51.2'
    ]
)
